"""A plugin for pylint to identify imports and functions which should not be used."""
from __future__ import (absolute_import, division, print_function)

__metaclass__ = type

import astroid

from pylint.checkers import BaseChecker
from pylint.interfaces import IAstroidChecker


class BlacklistEntry(object):
    """Defines a import blacklist entry."""
    def __init__(self, alternative, modules_only=False, names=None, ignore_paths=None):
        """
        :type alternative: str
        :type modules_only: bool
        :type names: tuple[str] | None
        :type ignore_paths: tuple[str] | None
        """
        self.alternative = alternative
        self.modules_only = modules_only
        self.names = set(names) if names else None
        self.ignore_paths = ignore_paths

    def applies_to(self, path, name=None):
        """
        :type path: str
        :type name: str | None
        :rtype: bool
        """
        if self.names:
            if not name:
                return False

            if name not in self.names:
                return False

        if self.ignore_paths and any(path.endswith(ignore_path) for ignore_path in self.ignore_paths):
            return False

        if self.modules_only:
            return is_module_path(path)

        return True


def is_module_path(path):
    """
    :type path: str
    :rtype: bool
    """
    return '/lib/ansible/modules/' in path or '/lib/ansible/module_utils/' in path


class AnsibleBlacklistChecker(BaseChecker):
    """Checker for blacklisted imports and functions."""
    __implements__ = (IAstroidChecker,)

    name = 'blacklist'

    BAD_IMPORT = 'ansible-bad-import'
    BAD_IMPORT_FROM = 'ansible-bad-import-from'
    BAD_FUNCTION = 'ansible-bad-function'
    BAD_MODULE_IMPORT = 'ansible-bad-module-import'

    msgs = dict(
        E5101=('Import %s instead of %s',
               BAD_IMPORT,
               'Identifies imports which should not be used.'),
        E5102=('Import %s from %s instead of %s',
               BAD_IMPORT_FROM,
               'Identifies imports which should not be used.'),
        E5103=('Call %s instead of %s',
               BAD_FUNCTION,
               'Identifies functions which should not be used.'),
        E5104=('Import external package or ansible.module_utils not %s',
               BAD_MODULE_IMPORT,
               'Identifies imports which should not be used.'),
    )

    blacklist_imports = dict(
        # Additional imports that we may want to start checking:
        # boto=BlacklistEntry('boto3', modules_only=True),
        # requests=BlacklistEntry('ansible.module_utils.urls', modules_only=True),
        # urllib=BlacklistEntry('ansible.module_utils.urls', modules_only=True),

        # see https://docs.python.org/2/library/urllib2.html
        urllib2=BlacklistEntry('ansible.module_utils.urls',
                               ignore_paths=(
                                   '/lib/ansible/module_utils/urls.py',
                               )),

        # see https://docs.python.org/3.7/library/collections.abc.html
        collections=BlacklistEntry('ansible.module_utils.common._collections_compat',
                                   ignore_paths=(
                                       '/lib/ansible/module_utils/common/_collections_compat.py',
                                   ),
                                   names=(
                                       'MappingView',
                                       'ItemsView',
                                       'KeysView',
                                       'ValuesView',
                                       'Mapping', 'MutableMapping',
                                       'Sequence', 'MutableSequence',
                                       'Set', 'MutableSet',
                                       'Container',
                                       'Hashable',
                                       'Sized',
                                       'Callable',
                                       'Iterable',
                                       'Iterator',
                                   )),
    )

    blacklist_functions = {
        # see https://docs.python.org/2/library/tempfile.html#tempfile.mktemp
        'tempfile.mktemp': BlacklistEntry('tempfile.mkstemp'),

        'sys.exit': BlacklistEntry('exit_json or fail_json',
                                   ignore_paths=(
                                       '/lib/ansible/module_utils/basic.py',
                                       '/lib/ansible/modules/utilities/logic/async_wrapper.py',
                                       '/lib/ansible/module_utils/common/removed.py',
                                   ),
                                   modules_only=True),
    }

    def visit_import(self, node):
        """
        :type node: astroid.node_classes.Import
        """
        for name in node.names:
            self._check_import(node, name[0])

    def visit_importfrom(self, node):
        """
        :type node: astroid.node_classes.ImportFrom
        """
        self._check_importfrom(node, node.modname, node.names)

    def visit_attribute(self, node):
        """
        :type node: astroid.node_classes.Attribute
        """
        last_child = node.last_child()

        # this is faster than using type inference and will catch the most common cases
        if not isinstance(last_child, astroid.node_classes.Name):
            return

        module = last_child.name

        entry = self.blacklist_imports.get(module)

        if entry and entry.names:
            if entry.applies_to(self.linter.current_file, node.attrname):
                self.add_message(self.BAD_IMPORT_FROM, args=(node.attrname, entry.alternative, module), node=node)

    def visit_call(self, node):
        """
        :type node: astroid.node_classes.Call
        """
        try:
            for i in node.func.inferred():
                func = None

                if isinstance(i, astroid.scoped_nodes.FunctionDef) and isinstance(i.parent, astroid.scoped_nodes.Module):
                    func = '%s.%s' % (i.parent.name, i.name)

                if not func:
                    continue

                entry = self.blacklist_functions.get(func)

                if entry and entry.applies_to(self.linter.current_file):
                    self.add_message(self.BAD_FUNCTION, args=(entry.alternative, func), node=node)
        except astroid.exceptions.InferenceError:
            pass

    def _check_import(self, node, modname):
        """
        :type node: astroid.node_classes.Import
        :type modname: str
        """
        self._check_module_import(node, modname)

        entry = self.blacklist_imports.get(modname)

        if not entry:
            return

        if entry.applies_to(self.linter.current_file):
            self.add_message(self.BAD_IMPORT, args=(entry.alternative, modname), node=node)

    def _check_importfrom(self, node, modname, names):
        """
        :type node: astroid.node_classes.ImportFrom
        :type modname: str
        :type names:  list[str[
        """
        self._check_module_import(node, modname)

        entry = self.blacklist_imports.get(modname)

        if not entry:
            return

        for name in names:
            if entry.applies_to(self.linter.current_file, name[0]):
                self.add_message(self.BAD_IMPORT_FROM, args=(name[0], entry.alternative, modname), node=node)

    def _check_module_import(self, node, modname):
        """
        :type node: astroid.node_classes.Import | astroid.node_classes.ImportFrom
        :type modname: str
        """
        if not is_module_path(self.linter.current_file):
            return

        if modname == 'ansible.module_utils' or modname.startswith('ansible.module_utils.'):
            return

        if modname == 'ansible' or modname.startswith('ansible.'):
            self.add_message(self.BAD_MODULE_IMPORT, args=(modname,), node=node)


def register(linter):
    """required method to auto register this checker """
    linter.register_checker(AnsibleBlacklistChecker(linter))



''' unit tests for Ansible module: na_ontap_vscan_on_demand_task '''

from __future__ import print_function
import json
import pytest

from units.compat import unittest
from units.compat.mock import patch, Mock
from ansible.module_utils import basic
from ansible.module_utils._text import to_bytes
import ansible.module_utils.netapp as netapp_utils

from ansible.modules.storage.netapp.na_ontap_vscan_on_demand_task \
    import NetAppOntapVscanOnDemandTask as onDemand_module  # module under test

if not netapp_utils.has_netapp_lib():
    pytestmark = pytest.mark.skip('skipping as missing required netapp_lib')


def set_module_args(args):
    """prepare arguments so that they will be picked up during module creation"""
    args = json.dumps({'ANSIBLE_MODULE_ARGS': args})
    basic._ANSIBLE_ARGS = to_bytes(args)  # pylint: disable=protected-access


class AnsibleExitJson(Exception):
    """Exception class to be raised by module.exit_json and caught by the test case"""
    pass


class AnsibleFailJson(Exception):
    """Exception class to be raised by module.fail_json and caught by the test case"""
    pass


def exit_json(*args, **kwargs):  # pylint: disable=unused-argument
    """function to patch over exit_json; package return data into an exception"""
    if 'changed' not in kwargs:
        kwargs['changed'] = False
    raise AnsibleExitJson(kwargs)


def fail_json(*args, **kwargs):  # pylint: disable=unused-argument
    """function to patch over fail_json; package return data into an exception"""
    kwargs['failed'] = True
    raise AnsibleFailJson(kwargs)


class MockONTAPConnection(object):
    ''' mock server connection to ONTAP host '''

    def __init__(self, kind=None, data=None):
        ''' save arguments '''
        self.kind = kind
        self.params = data
        self.xml_in = None
        self.xml_out = None

    def invoke_successfully(self, xml, enable_tunneling):  # pylint: disable=unused-argument
        ''' mock invoke_successfully returning xml data '''
        self.xml_in = xml
        if self.kind == 'task':
            xml = self.build_onDemand_pool_info(self.params)
        self.xml_out = xml
        return xml

    @staticmethod
    def build_onDemand_pool_info(onDemand_details):
        xml = netapp_utils.zapi.NaElement('xml')
        attributes = {
            'num-records': 1,
            'attributes-list': {
                'vscan-on-demand-task-info': {
                    'task-name': onDemand_details['task_name'],
                    'report-directory': onDemand_details['report_directory'],
                    'scan-paths': {
                        'string': onDemand_details['scan_paths']
                    }
                }
            }
        }
        xml.translate_struct(attributes)
        return xml


class TestMyModule(unittest.TestCase):
    ''' Unit tests for na_ontap_job_schedule '''

    def setUp(self):
        self.mock_module_helper = patch.multiple(basic.AnsibleModule,
                                                 exit_json=exit_json,
                                                 fail_json=fail_json)
        self.mock_module_helper.start()
        self.addCleanup(self.mock_module_helper.stop)
        self.mock_onDemand = {
            'state': 'present',
            'vserver': 'test_vserver',
            'report_directory': '/',
            'task_name': '/',
            'scan_paths': '/'
        }

    def mock_args(self):
        return {
            'state': self.mock_onDemand['state'],
            'vserver': self.mock_onDemand['vserver'],
            'report_directory': self.mock_onDemand['report_directory'],
            'task_name': self.mock_onDemand['task_name'],
            'scan_paths': self.mock_onDemand['scan_paths'],
            'hostname': 'test',
            'username': 'test_user',
            'password': 'test_pass!'
        }

    def get_demand_mock_object(self, kind=None):
        scanner_obj = onDemand_module()
        scanner_obj.asup_log_for_cserver = Mock(return_value=None)
        if kind is None:
            scanner_obj.server = MockONTAPConnection()
        else:
            scanner_obj.server = MockONTAPConnection(kind='task', data=self.mock_onDemand)
        return scanner_obj

    def test_module_fail_when_required_args_missing(self):
        ''' required arguments are reported as errors '''
        with pytest.raises(AnsibleFailJson) as exc:
            set_module_args({})
            onDemand_module()
        print('Info: %s' % exc.value.args[0]['msg'])

    def test_get_nonexistent_demand_task(self):
        set_module_args(self.mock_args())
        result = self.get_demand_mock_object().get_demand_task()
        assert not result

    def test_get_existing_demand_task(self):
        set_module_args(self.mock_args())
        result = self.get_demand_mock_object('task').get_demand_task()
        assert result

    def test_successfully_create(self):
        set_module_args(self.mock_args())
        with pytest.raises(AnsibleExitJson) as exc:
            self.get_demand_mock_object().apply()
        assert exc.value.args[0]['changed']

    def test_create_idempotency(self):
        set_module_args(self.mock_args())
        with pytest.raises(AnsibleExitJson) as exc:
            self.get_demand_mock_object('task').apply()
        assert not exc.value.args[0]['changed']

    def test_successfully_delete(self):
        data = self.mock_args()
        data['state'] = 'absent'
        set_module_args(data)
        with pytest.raises(AnsibleExitJson) as exc:
            self.get_demand_mock_object('task').apply()
        assert exc.value.args[0]['changed']

    def test_delete_idempotency(self):
        data = self.mock_args()
        data['state'] = 'absent'
        set_module_args(data)
        with pytest.raises(AnsibleExitJson) as exc:
            self.get_demand_mock_object().apply()
        assert not exc.value.args[0]['changed']



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2017, Flavien Chantelot (@Dorn-)
# Copyright: (c) 2018, Antoine Levy-Lambert (@antoinell)
# Copyright: (c) 2019, Andrew Klychkov (@Andersson007) <aaklychkov@mail.ru>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

ANSIBLE_METADATA = {
    'metadata_version': '1.1',
    'supported_by': 'community',
    'status': ['preview']
}

DOCUMENTATION = r'''
---
module: postgresql_tablespace
short_description: Add or remove PostgreSQL tablespaces from remote hosts
description:
- Adds or removes PostgreSQL tablespaces from remote hosts
  U(https://www.postgresql.org/docs/current/sql-createtablespace.html),
  U(https://www.postgresql.org/docs/current/manage-ag-tablespaces.html).
version_added: '2.8'
options:
  tablespace:
    description:
    - Name of the tablespace to add or remove.
    required: true
    type: str
    aliases:
    - name
  location:
    description:
    - Path to the tablespace directory in the file system.
    - Ensure that the location exists and has right privileges.
    type: path
    aliases:
    - path
  state:
    description:
    - Tablespace state.
    - I(state=present) implies the tablespace must be created if it doesn't exist.
    - I(state=absent) implies the tablespace must be removed if present.
      I(state=absent) is mutually exclusive with I(location), I(owner), i(set).
    - See the Notes section for information about check mode restrictions.
    type: str
    default: present
    choices: [ absent, present ]
  owner:
    description:
    - Name of the role to set as an owner of the tablespace.
    - If this option is not specified, the tablespace owner is a role that creates the tablespace.
    type: str
  set:
    description:
    - Dict of tablespace options to set. Supported from PostgreSQL 9.0.
    - For more information see U(https://www.postgresql.org/docs/current/sql-createtablespace.html).
    - When reset is passed as an option's value, if the option was set previously, it will be removed
      U(https://www.postgresql.org/docs/current/sql-altertablespace.html).
    type: dict
  rename_to:
    description:
    - New name of the tablespace.
    - The new name cannot begin with pg_, as such names are reserved for system tablespaces.
  session_role:
    description:
    - Switch to session_role after connecting. The specified session_role must
      be a role that the current login_user is a member of.
    - Permissions checking for SQL commands is carried out as though
      the session_role were the one that had logged in originally.
    type: str
  db:
    description:
    - Name of database to connect to and run queries against.
    type: str
    aliases:
    - login_db
notes:
- I(state=absent) and I(state=present) (the second one if the tablespace doesn't exist) do not
  support check mode because the corresponding PostgreSQL DROP and CREATE TABLESPACE commands
  can not be run inside the transaction block.
- The default authentication assumes that you are either logging in as or
  sudo'ing to the postgres account on the host.
- To avoid "Peer authentication failed for user postgres" error,
  use postgres user as a I(become_user).
- This module uses psycopg2, a Python PostgreSQL database adapter. You must
  ensure that psycopg2 is installed on the host before using this module.
- If the remote host is the PostgreSQL server (which is the default case), then
  PostgreSQL must also be installed on the remote host.
- For Ubuntu-based systems, install the postgresql, libpq-dev, and python-psycopg2 packages
  on the remote host before using this module.
requirements: [ psycopg2 ]
author:
- Flavien Chantelot (@Dorn-)
- Antoine Levy-Lambert (@antoinell)
- Andrew Klychkov (@Andersson007)
extends_documentation_fragment: postgres
'''

EXAMPLES = r'''
- name: Create a new tablespace called acme and set bob as an its owner
  postgresql_tablespace:
    name: acme
    owner: bob
    location: /data/foo

- name: Create a new tablespace called bar with tablespace options
  postgresql_tablespace:
    name: bar
    set:
      random_page_cost: 1
      seq_page_cost: 1

- name: Reset random_page_cost option
  postgresql_tablespace:
    name: bar
    set:
      random_page_cost: reset

- name: Rename the tablespace from bar to pcie_ssd
  postgresql_tablespace:
    name: bar
    rename_to: pcie_ssd

- name: Drop tablespace called bloat
  postgresql_tablespace:
    name: bloat
    state: absent
'''

RETURN = r'''
queries:
    description: List of queries that was tried to be executed.
    returned: always
    type: str
    sample: [ "CREATE TABLESPACE bar LOCATION '/incredible/ssd'" ]
tablespace:
    description: Tablespace name.
    returned: always
    type: str
    sample: 'ssd'
owner:
    description: Tablespace owner.
    returned: always
    type: str
    sample: 'Bob'
options:
    description: Tablespace options.
    returned: always
    type: dict
    sample: { 'random_page_cost': 1, 'seq_page_cost': 1 }
location:
    description: Path to the tablespace in the file system.
    returned: always
    type: str
    sample: '/incredible/fast/ssd'
newname:
    description: New tablespace name
    returned: if existent
    type: str
    sample: new_ssd
state:
    description: Tablespace state at the end of execution.
    returned: always
    type: str
    sample: 'present'
'''

try:
    from psycopg2 import __version__ as PSYCOPG2_VERSION
    from psycopg2.extras import DictCursor
    from psycopg2.extensions import ISOLATION_LEVEL_AUTOCOMMIT as AUTOCOMMIT
    from psycopg2.extensions import ISOLATION_LEVEL_READ_COMMITTED as READ_COMMITTED
except ImportError:
    # psycopg2 is checked by connect_to_db()
    # from ansible.module_utils.postgres
    pass

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.database import SQLParseError, pg_quote_identifier
from ansible.module_utils.postgres import connect_to_db, get_conn_params, postgres_common_argument_spec
from ansible.module_utils._text import to_native


class PgTablespace(object):
    def __init__(self, module, cursor, name):
        self.module = module
        self.cursor = cursor
        self.name = name
        self.exists = False
        self.owner = ''
        self.settings = {}
        self.location = ''
        self.executed_queries = []
        self.new_name = ''
        self.opt_not_supported = False
        # Collect info:
        self.get_info()

    def get_info(self):
        # Check that spcoptions exists:
        opt = self.__exec_sql("SELECT 1 FROM information_schema.columns "
                              "WHERE table_name = 'pg_tablespace' "
                              "AND column_name = 'spcoptions'", add_to_executed=False)

        # For 9.1 version and earlier:
        location = self.__exec_sql("SELECT 1 FROM information_schema.columns "
                                   "WHERE table_name = 'pg_tablespace' "
                                   "AND column_name = 'spclocation'", add_to_executed=False)
        if location:
            location = 'spclocation'
        else:
            location = 'pg_tablespace_location(t.oid)'

        if not opt:
            self.opt_not_supported = True
            query = ("SELECT r.rolname, (SELECT Null), %s "
                     "FROM pg_catalog.pg_tablespace AS t "
                     "JOIN pg_catalog.pg_roles AS r "
                     "ON t.spcowner = r.oid "
                     "WHERE t.spcname = '%s'" % (location, self.name))
        else:
            query = ("SELECT r.rolname, t.spcoptions, %s "
                     "FROM pg_catalog.pg_tablespace AS t "
                     "JOIN pg_catalog.pg_roles AS r "
                     "ON t.spcowner = r.oid "
                     "WHERE t.spcname = '%s'" % (location, self.name))

        res = self.__exec_sql(query, add_to_executed=False)

        if not res:
            self.exists = False
            return False

        if res[0][0]:
            self.exists = True
            self.owner = res[0][0]

            if res[0][1]:
                # Options exist:
                for i in res[0][1]:
                    i = i.split('=')
                    self.settings[i[0]] = i[1]

            if res[0][2]:
                # Location exists:
                self.location = res[0][2]

    def create(self, location):
        query = ("CREATE TABLESPACE %s LOCATION '%s'" % (pg_quote_identifier(self.name, 'database'), location))
        return self.__exec_sql(query, ddl=True)

    def drop(self):
        return self.__exec_sql("DROP TABLESPACE %s" % pg_quote_identifier(self.name, 'database'), ddl=True)

    def set_owner(self, new_owner):
        if new_owner == self.owner:
            return False

        query = "ALTER TABLESPACE %s OWNER TO %s" % (pg_quote_identifier(self.name, 'database'), new_owner)
        return self.__exec_sql(query, ddl=True)

    def rename(self, newname):
        query = "ALTER TABLESPACE %s RENAME TO %s" % (pg_quote_identifier(self.name, 'database'), newname)
        self.new_name = newname
        return self.__exec_sql(query, ddl=True)

    def set_settings(self, new_settings):
        # settings must be a dict {'key': 'value'}
        if self.opt_not_supported:
            return False

        changed = False

        # Apply new settings:
        for i in new_settings:
            if new_settings[i] == 'reset':
                if i in self.settings:
                    changed = self.__reset_setting(i)
                    self.settings[i] = None

            elif (i not in self.settings) or (str(new_settings[i]) != self.settings[i]):
                changed = self.__set_setting("%s = '%s'" % (i, new_settings[i]))

        return changed

    def __reset_setting(self, setting):
        query = "ALTER TABLESPACE %s RESET (%s)" % (pg_quote_identifier(self.name, 'database'), setting)
        return self.__exec_sql(query, ddl=True)

    def __set_setting(self, setting):
        query = "ALTER TABLESPACE %s SET (%s)" % (pg_quote_identifier(self.name, 'database'), setting)
        return self.__exec_sql(query, ddl=True)

    def __exec_sql(self, query, ddl=False, add_to_executed=True):
        try:
            self.cursor.execute(query)

            if add_to_executed:
                self.executed_queries.append(query)

            if not ddl:
                res = self.cursor.fetchall()
                return res
            return True
        except Exception as e:
            self.module.fail_json(msg="Cannot execute SQL '%s': %s" % (query, to_native(e)))
        return False


# ===========================================
# Module execution.
#


def main():
    argument_spec = postgres_common_argument_spec()
    argument_spec.update(
        tablespace=dict(type='str', aliases=['name']),
        state=dict(type='str', default="present", choices=["absent", "present"]),
        location=dict(type='path', aliases=['path']),
        owner=dict(type='str'),
        set=dict(type='dict'),
        rename_to=dict(type='str'),
        db=dict(type='str', aliases=['login_db']),
        session_role=dict(type='str'),
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        mutually_exclusive=(('positional_args', 'named_args'),),
        supports_check_mode=True,
    )

    tablespace = module.params["tablespace"]
    state = module.params["state"]
    location = module.params["location"]
    owner = module.params["owner"]
    rename_to = module.params["rename_to"]
    settings = module.params["set"]

    if state == 'absent' and (location or owner or rename_to or settings):
        module.fail_json(msg="state=absent is mutually exclusive location, "
                             "owner, rename_to, and set")

    conn_params = get_conn_params(module, module.params, warn_db_default=False)
    db_connection = connect_to_db(module, conn_params, autocommit=True)
    cursor = db_connection.cursor(cursor_factory=DictCursor)

    # Change autocommit to False if check_mode:
    if module.check_mode:
        if PSYCOPG2_VERSION >= '2.4.2':
            db_connection.set_session(autocommit=False)
        else:
            db_connection.set_isolation_level(READ_COMMITTED)

    # Set defaults:
    autocommit = False
    changed = False

    ##############
    # Create PgTablespace object and do main job:
    tblspace = PgTablespace(module, cursor, tablespace)

    # If tablespace exists with different location, exit:
    if tblspace.exists and location and location != tblspace.location:
        module.fail_json(msg="Tablespace '%s' exists with different location '%s'" % (tblspace.name, tblspace.location))

    # Create new tablespace:
    if not tblspace.exists and state == 'present':
        if rename_to:
            module.fail_json(msg="Tablespace %s does not exist, nothing to rename" % tablespace)

        if not location:
            module.fail_json(msg="'location' parameter must be passed with "
                                 "state=present if the tablespace doesn't exist")

        # Because CREATE TABLESPACE can not be run inside the transaction block:
        autocommit = True
        if PSYCOPG2_VERSION >= '2.4.2':
            db_connection.set_session(autocommit=True)
        else:
            db_connection.set_isolation_level(AUTOCOMMIT)

        changed = tblspace.create(location)

    # Drop non-existing tablespace:
    elif not tblspace.exists and state == 'absent':
        # Nothing to do:
        module.fail_json(msg="Tries to drop nonexistent tablespace '%s'" % tblspace.name)

    # Drop existing tablespace:
    elif tblspace.exists and state == 'absent':
        # Because DROP TABLESPACE can not be run inside the transaction block:
        autocommit = True
        if PSYCOPG2_VERSION >= '2.4.2':
            db_connection.set_session(autocommit=True)
        else:
            db_connection.set_isolation_level(AUTOCOMMIT)

        changed = tblspace.drop()

    # Rename tablespace:
    elif tblspace.exists and rename_to:
        if tblspace.name != rename_to:
            changed = tblspace.rename(rename_to)

    if state == 'present':
        # Refresh information:
        tblspace.get_info()

    # Change owner and settings:
    if state == 'present' and tblspace.exists:
        if owner:
            changed = tblspace.set_owner(owner)

        if settings:
            changed = tblspace.set_settings(settings)

        tblspace.get_info()

    # Rollback if it's possible and check_mode:
    if not autocommit:
        if module.check_mode:
            db_connection.rollback()
        else:
            db_connection.commit()

    cursor.close()
    db_connection.close()

    # Make return values:
    kw = dict(
        changed=changed,
        state='present',
        tablespace=tblspace.name,
        owner=tblspace.owner,
        queries=tblspace.executed_queries,
        options=tblspace.settings,
        location=tblspace.location,
    )

    if state == 'present':
        kw['state'] = 'present'

        if tblspace.new_name:
            kw['newname'] = tblspace.new_name

    elif state == 'absent':
        kw['state'] = 'absent'

    module.exit_json(**kw)


if __name__ == '__main__':
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2017, Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

# This is a virtual module that is entirely implemented as an action plugin and runs on the controller

from __future__ import absolute_import, division, print_function
__metaclass__ = type

ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['stableinterface'],
                    'supported_by': 'core'}

DOCUMENTATION = r'''
---
module: template
version_added: historical
short_description: Template a file out to a remote server
description:
- Templates are processed by the L(Jinja2 templating language,http://jinja.pocoo.org/docs/).
- Documentation on the template formatting can be found in the
  L(Template Designer Documentation,http://jinja.pocoo.org/docs/templates/).
- Additional variables listed below can be used in templates.
- C(ansible_managed) (configurable via the C(defaults) section of C(ansible.cfg)) contains a string which can be used to
  describe the template name, host, modification time of the template file and the owner uid.
- C(template_host) contains the node name of the template's machine.
- C(template_uid) is the numeric user id of the owner.
- C(template_path) is the path of the template.
- C(template_fullpath) is the absolute path of the template.
- C(template_destpath) is the path of the template on the remote system (added in 2.8).
- C(template_run_date) is the date that the template was rendered.
options:
  src:
    description:
    - Path of a Jinja2 formatted template on the Ansible controller.
    - This can be a relative or an absolute path.
    type: path
    required: yes
  dest:
    description:
    - Location to render the template to on the remote machine.
    type: path
    required: yes
  backup:
    description:
    - Determine whether a backup should be created.
    - When set to C(yes), create a backup file including the timestamp information
      so you can get the original file back if you somehow clobbered it incorrectly.
    type: bool
    default: no
  newline_sequence:
    description:
    - Specify the newline sequence to use for templating files.
    type: str
    choices: [ '\n', '\r', '\r\n' ]
    default: '\n'
    version_added: '2.4'
  block_start_string:
    description:
    - The string marking the beginning of a block.
    type: str
    default: '{%'
    version_added: '2.4'
  block_end_string:
    description:
    - The string marking the end of a block.
    type: str
    default: '%}'
    version_added: '2.4'
  variable_start_string:
    description:
    - The string marking the beginning of a print statement.
    type: str
    default: '{{'
    version_added: '2.4'
  variable_end_string:
    description:
    - The string marking the end of a print statement.
    type: str
    default: '}}'
    version_added: '2.4'
  trim_blocks:
    description:
    - Determine when newlines should be removed from blocks.
    - When set to C(yes) the first newline after a block is removed (block, not variable tag!).
    type: bool
    default: yes
    version_added: '2.4'
  lstrip_blocks:
    description:
    - Determine when leading spaces and tabs should be stripped.
    - When set to C(yes) leading spaces and tabs are stripped from the start of a line to a block.
    - This functionality requires Jinja 2.7 or newer.
    type: bool
    default: no
    version_added: '2.6'
  force:
    description:
    - Determine when the file is being transferred if the destination already exists.
    - When set to C(yes), replace the remote file when contents are different than the source.
    - When set to C(no), the file will only be transferred if the destination does not exist.
    type: bool
    default: yes
  follow:
    description:
    - Determine whether symbolic links should be followed.
    - When set to C(yes) symbolic links will be followed, if they exist.
    - When set to C(no) symbolic links will not be followed.
    - Previous to Ansible 2.4, this was hardcoded as C(yes).
    type: bool
    default: no
    version_added: '2.4'
  output_encoding:
    description:
    - Overrides the encoding used to write the template file defined by C(dest).
    - It defaults to C(utf-8), but any encoding supported by python can be used.
    - The source template file must always be encoded using C(utf-8), for homogeneity.
    type: str
    default: utf-8
    version_added: '2.7'
notes:
- Including a string that uses a date in the template will result in the template being marked 'changed' each time.
- Since Ansible 0.9, templates are loaded with C(trim_blocks=True).
- >
  Also, you can override jinja2 settings by adding a special header to template file.
  i.e. C(#jinja2:variable_start_string:'[%', variable_end_string:'%]', trim_blocks: False)
  which changes the variable interpolation markers to C([% var %]) instead of C({{ var }}).
  This is the best way to prevent evaluation of things that look like, but should not be Jinja2.
- Using raw/endraw in Jinja2 will not work as you expect because templates in Ansible are recursively
  evaluated.
- You can use the M(copy) module with the C(content:) option if you prefer the template inline,
  as part of the playbook.
- For Windows you can use M(win_template) which uses '\\r\\n' as C(newline_sequence) by default.
seealso:
- module: copy
- module: win_copy
- module: win_template
author:
- Ansible Core Team
- Michael DeHaan
extends_documentation_fragment:
- files
- validate
'''

EXAMPLES = r'''
- name: Template a file to /etc/files.conf
  template:
    src: /mytemplates/foo.j2
    dest: /etc/file.conf
    owner: bin
    group: wheel
    mode: '0644'

- name: Template a file, using symbolic modes (equivalent to 0644)
  template:
    src: /mytemplates/foo.j2
    dest: /etc/file.conf
    owner: bin
    group: wheel
    mode: u=rw,g=r,o=r

- name: Copy a version of named.conf that is dependent on the OS. setype obtained by doing ls -Z /etc/named.conf on original file
  template:
    src: named.conf_{{ ansible_os_family}}.j2
    dest: /etc/named.conf
    group: named
    setype: named_conf_t
    mode: 0640

- name: Create a DOS-style text file from a template
  template:
    src: config.ini.j2
    dest: /share/windows/config.ini
    newline_sequence: '\r\n'

- name: Copy a new sudoers file into place, after passing validation with visudo
  template:
    src: /mine/sudoers
    dest: /etc/sudoers
    validate: /usr/sbin/visudo -cf %s

- name: Update sshd configuration safely, avoid locking yourself out
  template:
    src: etc/ssh/sshd_config.j2
    dest: /etc/ssh/sshd_config
    owner: root
    group: root
    mode: '0600'
    validate: /usr/sbin/sshd -t -f %s
    backup: yes
'''



# -*- coding: utf-8 -*-
# Copyright (c) 2019 Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

import pytest

from ansible.module_utils.common.parameters import list_no_log_values


@pytest.fixture
def params():
    return {
        'secret': 'undercookwovennativity',
        'other_secret': 'cautious-slate-makeshift',
        'state': 'present',
        'value': 5,
    }


def test_list_no_log_values(params):
    argument_spec = {
        'secret': {'type': 'str', 'no_log': True},
        'other_secret': {'type': 'str', 'no_log': True},
        'state': {'type': 'str'},
        'value': {'type': 'int'},
    }
    result = set(('undercookwovennativity', 'cautious-slate-makeshift'))
    assert result == list_no_log_values(argument_spec, params)


def test_list_no_log_values_no_secrets(params):
    argument_spec = {
        'other_secret': {'type': 'str', 'no_log': False},
        'state': {'type': 'str'},
        'value': {'type': 'int'},
    }
    result = set()
    assert result == list_no_log_values(argument_spec, params)



#!/usr/bin/python
from __future__ import (absolute_import, division, print_function)
# Copyright 2019 Fortinet, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# the lib use python logging can get it if the following is set in your
# Ansible config.

__metaclass__ = type

ANSIBLE_METADATA = {'status': ['preview'],
                    'supported_by': 'community',
                    'metadata_version': '1.1'}

DOCUMENTATION = '''
---
module: fortios_log_syslogd_setting
short_description: Global settings for remote syslog server in Fortinet's FortiOS and FortiGate.
description:
    - This module is able to configure a FortiGate or FortiOS by allowing the
      user to set and modify log_syslogd feature and setting category.
      Examples include all parameters and values need to be adjusted to datasources before usage.
      Tested with FOS v6.0.2
version_added: "2.8"
author:
    - Miguel Angel Munoz (@mamunozgonzalez)
    - Nicolas Thomas (@thomnico)
notes:
    - Requires fortiosapi library developed by Fortinet
    - Run as a local_action in your playbook
requirements:
    - fortiosapi>=0.9.8
options:
    host:
       description:
            - FortiOS or FortiGate ip address.
       required: true
    username:
        description:
            - FortiOS or FortiGate username.
        required: true
    password:
        description:
            - FortiOS or FortiGate password.
        default: ""
    vdom:
        description:
            - Virtual domain, among those defined previously. A vdom is a
              virtual instance of the FortiGate that can be configured and
              used as a different unit.
        default: root
    https:
        description:
            - Indicates if the requests towards FortiGate must use HTTPS
              protocol
        type: bool
        default: true
    log_syslogd_setting:
        description:
            - Global settings for remote syslog server.
        default: null
        suboptions:
            certificate:
                description:
                    - Certificate used to communicate with Syslog server. Source certificate.local.name.
            custom-field-name:
                description:
                    - Custom field name for CEF format logging.
                suboptions:
                    custom:
                        description:
                            - Field custom name.
                    id:
                        description:
                            - Entry ID.
                        required: true
                    name:
                        description:
                            - Field name.
            enc-algorithm:
                description:
                    - Enable/disable reliable syslogging with TLS encryption.
                choices:
                    - high-medium
                    - high
                    - low
                    - disable
            facility:
                description:
                    - Remote syslog facility.
                choices:
                    - kernel
                    - user
                    - mail
                    - daemon
                    - auth
                    - syslog
                    - lpr
                    - news
                    - uucp
                    - cron
                    - authpriv
                    - ftp
                    - ntp
                    - audit
                    - alert
                    - clock
                    - local0
                    - local1
                    - local2
                    - local3
                    - local4
                    - local5
                    - local6
                    - local7
            format:
                description:
                    - Log format.
                choices:
                    - default
                    - csv
                    - cef
            mode:
                description:
                    - Remote syslog logging over UDP/Reliable TCP.
                choices:
                    - udp
                    - legacy-reliable
                    - reliable
            port:
                description:
                    - Server listen port.
            server:
                description:
                    - Address of remote syslog server.
            source-ip:
                description:
                    - Source IP address of syslog.
            status:
                description:
                    - Enable/disable remote syslog logging.
                choices:
                    - enable
                    - disable
'''

EXAMPLES = '''
- hosts: localhost
  vars:
   host: "192.168.122.40"
   username: "admin"
   password: ""
   vdom: "root"
  tasks:
  - name: Global settings for remote syslog server.
    fortios_log_syslogd_setting:
      host:  "{{ host }}"
      username: "{{ username }}"
      password: "{{ password }}"
      vdom:  "{{ vdom }}"
      https: "False"
      log_syslogd_setting:
        certificate: "<your_own_value> (source certificate.local.name)"
        custom-field-name:
         -
            custom: "<your_own_value>"
            id:  "6"
            name: "default_name_7"
        enc-algorithm: "high-medium"
        facility: "kernel"
        format: "default"
        mode: "udp"
        port: "12"
        server: "192.168.100.40"
        source-ip: "84.230.14.43"
        status: "enable"
'''

RETURN = '''
build:
  description: Build number of the fortigate image
  returned: always
  type: str
  sample: '1547'
http_method:
  description: Last method used to provision the content into FortiGate
  returned: always
  type: str
  sample: 'PUT'
http_status:
  description: Last result given by FortiGate on last operation applied
  returned: always
  type: str
  sample: "200"
mkey:
  description: Master key (id) used in the last call to FortiGate
  returned: success
  type: str
  sample: "id"
name:
  description: Name of the table used to fulfill the request
  returned: always
  type: str
  sample: "urlfilter"
path:
  description: Path of the table used to fulfill the request
  returned: always
  type: str
  sample: "webfilter"
revision:
  description: Internal revision number
  returned: always
  type: str
  sample: "17.0.2.10658"
serial:
  description: Serial number of the unit
  returned: always
  type: str
  sample: "FGVMEVYYQT3AB5352"
status:
  description: Indication of the operation's result
  returned: always
  type: str
  sample: "success"
vdom:
  description: Virtual domain used
  returned: always
  type: str
  sample: "root"
version:
  description: Version of the FortiGate
  returned: always
  type: str
  sample: "v5.6.3"

'''

from ansible.module_utils.basic import AnsibleModule

fos = None


def login(data):
    host = data['host']
    username = data['username']
    password = data['password']

    fos.debug('on')
    if 'https' in data and not data['https']:
        fos.https('off')
    else:
        fos.https('on')

    fos.login(host, username, password)


def filter_log_syslogd_setting_data(json):
    option_list = ['certificate', 'custom-field-name', 'enc-algorithm',
                   'facility', 'format', 'mode',
                   'port', 'server', 'source-ip',
                   'status']
    dictionary = {}

    for attribute in option_list:
        if attribute in json and json[attribute] is not None:
            dictionary[attribute] = json[attribute]

    return dictionary


def flatten_multilists_attributes(data):
    multilist_attrs = []

    for attr in multilist_attrs:
        try:
            path = "data['" + "']['".join(elem for elem in attr) + "']"
            current_val = eval(path)
            flattened_val = ' '.join(elem for elem in current_val)
            exec(path + '= flattened_val')
        except BaseException:
            pass

    return data


def log_syslogd_setting(data, fos):
    vdom = data['vdom']
    log_syslogd_setting_data = data['log_syslogd_setting']
    flattened_data = flatten_multilists_attributes(log_syslogd_setting_data)
    filtered_data = filter_log_syslogd_setting_data(flattened_data)
    return fos.set('log.syslogd',
                   'setting',
                   data=filtered_data,
                   vdom=vdom)


def fortios_log_syslogd(data, fos):
    login(data)

    if data['log_syslogd_setting']:
        resp = log_syslogd_setting(data, fos)

    fos.logout()
    return not resp['status'] == "success", resp['status'] == "success", resp


def main():
    fields = {
        "host": {"required": True, "type": "str"},
        "username": {"required": True, "type": "str"},
        "password": {"required": False, "type": "str", "no_log": True},
        "vdom": {"required": False, "type": "str", "default": "root"},
        "https": {"required": False, "type": "bool", "default": True},
        "log_syslogd_setting": {
            "required": False, "type": "dict",
            "options": {
                "certificate": {"required": False, "type": "str"},
                "custom-field-name": {"required": False, "type": "list",
                                      "options": {
                                          "custom": {"required": False, "type": "str"},
                                          "id": {"required": True, "type": "int"},
                                          "name": {"required": False, "type": "str"}
                                      }},
                "enc-algorithm": {"required": False, "type": "str",
                                  "choices": ["high-medium", "high", "low",
                                              "disable"]},
                "facility": {"required": False, "type": "str",
                             "choices": ["kernel", "user", "mail",
                                         "daemon", "auth", "syslog",
                                         "lpr", "news", "uucp",
                                         "cron", "authpriv", "ftp",
                                         "ntp", "audit", "alert",
                                         "clock", "local0", "local1",
                                         "local2", "local3", "local4",
                                         "local5", "local6", "local7"]},
                "format": {"required": False, "type": "str",
                           "choices": ["default", "csv", "cef"]},
                "mode": {"required": False, "type": "str",
                         "choices": ["udp", "legacy-reliable", "reliable"]},
                "port": {"required": False, "type": "int"},
                "server": {"required": False, "type": "str"},
                "source-ip": {"required": False, "type": "str"},
                "status": {"required": False, "type": "str",
                           "choices": ["enable", "disable"]}

            }
        }
    }

    module = AnsibleModule(argument_spec=fields,
                           supports_check_mode=False)
    try:
        from fortiosapi import FortiOSAPI
    except ImportError:
        module.fail_json(msg="fortiosapi module is required")

    global fos
    fos = FortiOSAPI()

    is_error, has_changed, result = fortios_log_syslogd(module.params, fos)

    if not is_error:
        module.exit_json(changed=has_changed, meta=result)
    else:
        module.fail_json(msg="Error in repo", meta=result)


if __name__ == '__main__':
    main()



# (c) 2018, NetApp, Inc
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

''' unit test template for ONTAP Ansible module '''

from __future__ import print_function
import json
import pytest

from units.compat import unittest
from units.compat.mock import patch, Mock
from ansible.module_utils import basic
from ansible.module_utils._text import to_bytes
import ansible.module_utils.netapp as netapp_utils

from ansible.modules.storage.netapp.na_ontap_net_ifgrp \
    import NetAppOntapIfGrp as ifgrp_module  # module under test

if not netapp_utils.has_netapp_lib():
    pytestmark = pytest.mark.skip('skipping as missing required netapp_lib')


def set_module_args(args):
    """prepare arguments so that they will be picked up during module creation"""
    args = json.dumps({'ANSIBLE_MODULE_ARGS': args})
    basic._ANSIBLE_ARGS = to_bytes(args)  # pylint: disable=protected-access


class AnsibleExitJson(Exception):
    """Exception class to be raised by module.exit_json and caught by the test case"""
    pass


class AnsibleFailJson(Exception):
    """Exception class to be raised by module.fail_json and caught by the test case"""
    pass


def exit_json(*args, **kwargs):  # pylint: disable=unused-argument
    """function to patch over exit_json; package return data into an exception"""
    if 'changed' not in kwargs:
        kwargs['changed'] = False
    raise AnsibleExitJson(kwargs)


def fail_json(*args, **kwargs):  # pylint: disable=unused-argument
    """function to patch over fail_json; package return data into an exception"""
    kwargs['failed'] = True
    raise AnsibleFailJson(kwargs)


class MockONTAPConnection(object):
    ''' mock server connection to ONTAP host '''

    def __init__(self, kind=None, data=None):
        ''' save arguments '''
        self.kind = kind
        self.params = data
        self.xml_in = None
        self.xml_out = None

    def invoke_successfully(self, xml, enable_tunneling):  # pylint: disable=unused-argument
        ''' mock invoke_successfully returning xml data '''
        self.xml_in = xml
        if self.kind == 'ifgrp':
            xml = self.build_ifgrp_info(self.params)
        elif self.kind == 'ifgrp-ports':
            xml = self.build_ifgrp_ports_info(self.params)
        elif self.kind == 'ifgrp-fail':
            raise netapp_utils.zapi.NaApiError(code='TEST', message="This exception is from the unit test")
        self.xml_out = xml
        return xml

    @staticmethod
    def build_ifgrp_info(ifgrp_details):
        ''' build xml data for ifgrp-info '''
        xml = netapp_utils.zapi.NaElement('xml')
        attributes = {
            'num-records': 1,
            'attributes-list': {
                'net-port-info': {
                    'port': ifgrp_details['name'],
                    'ifgrp-distribution-function': 'mac',
                    'ifgrp-mode': ifgrp_details['mode'],
                    'node': ifgrp_details['node']
                }
            }
        }
        xml.translate_struct(attributes)
        return xml

    @staticmethod
    def build_ifgrp_ports_info(data):
        ''' build xml data for ifgrp-ports '''
        xml = netapp_utils.zapi.NaElement('xml')
        attributes = {
            'attributes': {
                'net-ifgrp-info': {
                    'ports': [
                        {'lif-bindable': data['ports'][0]},
                        {'lif-bindable': data['ports'][1]},
                        {'lif-bindable': data['ports'][2]}
                    ]
                }
            }
        }
        xml.translate_struct(attributes)
        return xml


class TestMyModule(unittest.TestCase):
    ''' a group of related Unit Tests '''

    def setUp(self):
        self.mock_module_helper = patch.multiple(basic.AnsibleModule,
                                                 exit_json=exit_json,
                                                 fail_json=fail_json)
        self.mock_module_helper.start()
        self.addCleanup(self.mock_module_helper.stop)
        self.mock_ifgrp = {
            'name': 'test',
            'port': 'a1',
            'node': 'test_vserver',
            'mode': 'something'
        }

    def mock_args(self):
        return {
            'name': self.mock_ifgrp['name'],
            'distribution_function': 'mac',
            'ports': [self.mock_ifgrp['port']],
            'node': self.mock_ifgrp['node'],
            'mode': self.mock_ifgrp['mode'],
            'hostname': 'test',
            'username': 'test_user',
            'password': 'test_pass!'
        }

    def get_ifgrp_mock_object(self, kind=None, data=None):
        """
        Helper method to return an na_ontap_net_ifgrp object
        :param kind: passes this param to MockONTAPConnection()
        :return: na_ontap_net_ifgrp object
        """
        obj = ifgrp_module()
        obj.autosupport_log = Mock(return_value=None)
        if data is None:
            data = self.mock_ifgrp
        obj.server = MockONTAPConnection(kind=kind, data=data)
        return obj

    def test_module_fail_when_required_args_missing(self):
        ''' required arguments are reported as errors '''
        with pytest.raises(AnsibleFailJson) as exc:
            set_module_args({})
            ifgrp_module()
        print('Info: %s' % exc.value.args[0]['msg'])

    def test_get_nonexistent_ifgrp(self):
        ''' Test if get_ifgrp returns None for non-existent ifgrp '''
        set_module_args(self.mock_args())
        result = self.get_ifgrp_mock_object().get_if_grp()
        assert result is None

    def test_get_existing_ifgrp(self):
        ''' Test if get_ifgrp returns details for existing ifgrp '''
        set_module_args(self.mock_args())
        result = self.get_ifgrp_mock_object('ifgrp').get_if_grp()
        assert result['name'] == self.mock_ifgrp['name']

    def test_successful_create(self):
        ''' Test successful create '''
        data = self.mock_args()
        set_module_args(data)
        with pytest.raises(AnsibleExitJson) as exc:
            self.get_ifgrp_mock_object().apply()
        assert exc.value.args[0]['changed']

    def test_successful_delete(self):
        ''' Test delete existing volume '''
        data = self.mock_args()
        data['state'] = 'absent'
        set_module_args(data)
        with pytest.raises(AnsibleExitJson) as exc:
            self.get_ifgrp_mock_object('ifgrp').apply()
        assert exc.value.args[0]['changed']

    def test_successful_modify(self):
        ''' Test delete existing volume '''
        data = self.mock_args()
        data['ports'] = ['1', '2', '3']
        set_module_args(data)
        with pytest.raises(AnsibleExitJson) as exc:
            self.get_ifgrp_mock_object('ifgrp').apply()
        assert exc.value.args[0]['changed']

    @patch('ansible.modules.storage.netapp.na_ontap_net_ifgrp.NetAppOntapIfGrp.get_if_grp')
    @patch('ansible.modules.storage.netapp.na_ontap_net_ifgrp.NetAppOntapIfGrp.create_if_grp')
    def test_create_called(self, create_ifgrp, get_ifgrp):
        data = self.mock_args()
        set_module_args(data)
        get_ifgrp.return_value = None
        with pytest.raises(AnsibleExitJson) as exc:
            self.get_ifgrp_mock_object().apply()
        get_ifgrp.assert_called_with()
        create_ifgrp.assert_called_with()

    @patch('ansible.modules.storage.netapp.na_ontap_net_ifgrp.NetAppOntapIfGrp.add_port_to_if_grp')
    def test_if_ports_are_added_after_create(self, add_ports):
        ''' Test successful create '''
        data = self.mock_args()
        set_module_args(data)
        self.get_ifgrp_mock_object().create_if_grp()
        add_ports.assert_called_with('a1')

    @patch('ansible.modules.storage.netapp.na_ontap_net_ifgrp.NetAppOntapIfGrp.get_if_grp')
    @patch('ansible.modules.storage.netapp.na_ontap_net_ifgrp.NetAppOntapIfGrp.delete_if_grp')
    def test_delete_called(self, delete_ifgrp, get_ifgrp):
        data = self.mock_args()
        data['state'] = 'absent'
        set_module_args(data)
        get_ifgrp.return_value = Mock()
        with pytest.raises(AnsibleExitJson) as exc:
            self.get_ifgrp_mock_object().apply()
        get_ifgrp.assert_called_with()
        delete_ifgrp.assert_called_with()

    def test_get_return_value(self):
        data = self.mock_args()
        set_module_args(data)
        result = self.get_ifgrp_mock_object('ifgrp').get_if_grp()
        assert result['name'] == data['name']
        assert result['mode'] == data['mode']
        assert result['node'] == data['node']

    def test_get_ports_list(self):
        data = self.mock_args()
        data['ports'] = ['e0a', 'e0b', 'e0c']
        set_module_args(data)
        result = self.get_ifgrp_mock_object('ifgrp-ports', data).get_if_grp_ports()
        assert result['ports'] == data['ports']

    def test_add_port_packet(self):
        data = self.mock_args()
        set_module_args(data)
        obj = self.get_ifgrp_mock_object('ifgrp')
        obj.add_port_to_if_grp('addme')
        assert obj.server.xml_in['port'] == 'addme'

    @patch('ansible.modules.storage.netapp.na_ontap_net_ifgrp.NetAppOntapIfGrp.remove_port_to_if_grp')
    @patch('ansible.modules.storage.netapp.na_ontap_net_ifgrp.NetAppOntapIfGrp.add_port_to_if_grp')
    def test_modify_ports_calls_remove_existing_ports(self, add_port, remove_port):
        ''' Test if already existing ports are not being added again '''
        data = self.mock_args()
        data['ports'] = ['1', '2']
        set_module_args(data)
        self.get_ifgrp_mock_object('ifgrp').modify_ports(current_ports=['1', '2', '3'])
        assert remove_port.call_count == 1
        assert add_port.call_count == 0

    @patch('ansible.modules.storage.netapp.na_ontap_net_ifgrp.NetAppOntapIfGrp.remove_port_to_if_grp')
    @patch('ansible.modules.storage.netapp.na_ontap_net_ifgrp.NetAppOntapIfGrp.add_port_to_if_grp')
    def test_modify_ports_calls_add_new_ports(self, add_port, remove_port):
        ''' Test new ports are added '''
        data = self.mock_args()
        data['ports'] = ['1', '2', '3', '4']
        set_module_args(data)
        self.get_ifgrp_mock_object('ifgrp').modify_ports(current_ports=['1', '2'])
        assert remove_port.call_count == 0
        assert add_port.call_count == 2

    def test_get_ports_returns_none(self):
        set_module_args(self.mock_args())
        result = self.get_ifgrp_mock_object().get_if_grp_ports()
        assert result['ports'] == []
        result = self.get_ifgrp_mock_object().get_if_grp()
        assert result is None

    def test_if_all_methods_catch_exception(self):
        set_module_args(self.mock_args())
        with pytest.raises(AnsibleFailJson) as exc:
            self.get_ifgrp_mock_object('ifgrp-fail').get_if_grp()
        assert 'Error getting if_group test' in exc.value.args[0]['msg']
        with pytest.raises(AnsibleFailJson) as exc:
            self.get_ifgrp_mock_object('ifgrp-fail').create_if_grp()
        assert 'Error creating if_group test' in exc.value.args[0]['msg']
        with pytest.raises(AnsibleFailJson) as exc:
            self.get_ifgrp_mock_object('ifgrp-fail').get_if_grp_ports()
        assert 'Error getting if_group ports test' in exc.value.args[0]['msg']
        with pytest.raises(AnsibleFailJson) as exc:
            self.get_ifgrp_mock_object('ifgrp-fail').add_port_to_if_grp('test-port')
        assert 'Error adding port test-port to if_group test' in exc.value.args[0]['msg']
        with pytest.raises(AnsibleFailJson) as exc:
            self.get_ifgrp_mock_object('ifgrp-fail').remove_port_to_if_grp('test-port')
        assert 'Error removing port test-port to if_group test' in exc.value.args[0]['msg']
        with pytest.raises(AnsibleFailJson) as exc:
            self.get_ifgrp_mock_object('ifgrp-fail').delete_if_grp()
        assert 'Error deleting if_group test' in exc.value.args[0]['msg']



#!/usr/bin/python
# Copyright: (c) 2018, Pluribus Networks
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'community'}


DOCUMENTATION = """
---
module: pn_dscp_map_pri_map
author: "Pluribus Networks (@rajaspachipulusu17)"
version_added: "2.8"
short_description: CLI command to modify dscp-map-pri-map
description:
  - This module can be used to update priority mappings in tables.
options:
  pn_cliswitch:
    description:
      - Target switch to run the CLI on.
    required: False
    type: str
  state:
    description:
      - State the action to perform. Use C(update) to modify
        the dscp-map-pri-map.
    required: True
    type: str
    choices: ['update']
  pn_pri:
    description:
      - CoS priority.
    required: False
    type: str
  pn_name:
    description:
      - Name for the DSCP map.
    required: False
    type: str
  pn_dsmap:
    description:
      - DSCP value(s).
    required: False
    type: str
"""

EXAMPLES = """
- name: dscp map pri map modify
  pn_dscp_map_pri_map:
    pn_cliswitch: 'sw01'
    state: 'update'
    pn_name: 'foo'
    pn_pri: '0'
    pn_dsmap: '40'

- name: dscp map pri map modify
  pn_dscp_map_pri_map:
    pn_cliswitch: 'sw01'
    state: 'update'
    pn_name: 'foo'
    pn_pri: '1'
    pn_dsmap: '8,10,12,14'
"""

RETURN = """
command:
  description: the CLI command run on the target node.
  returned: always
  type: str
stdout:
  description: set of responses from the dscp-map-pri-map command.
  returned: always
  type: list
stderr:
  description: set of error responses from the dscp-map-pri-map command.
  returned: on error
  type: list
changed:
  description: indicates whether the CLI caused changes on the target.
  returned: always
  type: bool
"""

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.network.netvisor.pn_nvos import pn_cli, run_cli
from ansible.module_utils.network.netvisor.netvisor import run_commands


def check_cli(module, cli):
    """
    This method checks for idempotency using the dscp-map-show name command.
    If a user with given name exists, return True else False.
    :param module: The Ansible module to fetch input parameters
    :param cli: The CLI string
    """
    name = module.params['pn_name']

    cli += ' dscp-map-show name %s format name no-show-headers' % name
    out = run_commands(module, cli)[1]

    out = out.split()

    return True if name in out[-1] else False


def main():
    """ This section is for arguments parsing """

    state_map = dict(
        update='dscp-map-pri-map-modify'
    )
    module = AnsibleModule(
        argument_spec=dict(
            pn_cliswitch=dict(required=False, type='str'),
            state=dict(required=True, type='str',
                       choices=state_map.keys()),
            pn_pri=dict(required=False, type='str'),
            pn_name=dict(required=False, type='str'),
            pn_dsmap=dict(required=False, type='str'),
        ),
        required_if=(
            ['state', 'update', ['pn_name', 'pn_pri']],
        )
    )

    # Accessing the arguments
    cliswitch = module.params['pn_cliswitch']
    state = module.params['state']
    pri = module.params['pn_pri']
    name = module.params['pn_name']
    dsmap = module.params['pn_dsmap']

    command = state_map[state]

    # Building the CLI command string
    cli = pn_cli(module, cliswitch)

    NAME_EXISTS = check_cli(module, cli)

    if command == 'dscp-map-pri-map-modify':
        if NAME_EXISTS is False:
            module.fail_json(
                failed=True,
                msg='Create dscp map with name %s before updating' % name
            )
        cli += ' %s ' % command
        if pri:
            cli += ' pri ' + pri
        if name:
            cli += ' name ' + name
        if dsmap:
            cli += ' dsmap ' + dsmap

    run_cli(module, cli, state_map)


if __name__ == '__main__':
    main()



#!/usr/bin/python
from __future__ import (absolute_import, division, print_function)
# Copyright 2018 Fortinet, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# the lib use python logging can get it if the following is set in your
# Ansible config.

__metaclass__ = type

ANSIBLE_METADATA = {'status': ['preview'],
                    'supported_by': 'community',
                    'metadata_version': '1.1'}

DOCUMENTATION = '''
---
module: fortios_firewall_address
short_description: Configure IPv4 addresses.
description:
    - This module is able to configure a FortiGate or FortiOS by
      allowing the user to configure firewall feature and address category.
      Examples includes all options and need to be adjusted to datasources before usage.
      Tested with FOS v6.0.2
version_added: "2.8"
author:
    - Miguel Angel Munoz (@mamunozgonzalez)
    - Nicolas Thomas (@thomnico)
notes:
    - Requires fortiosapi library developed by Fortinet
    - Run as a local_action in your playbook
requirements:
    - fortiosapi>=0.9.8
options:
    host:
       description:
            - FortiOS or FortiGate ip address.
       required: true
    username:
        description:
            - FortiOS or FortiGate username.
        required: true
    password:
        description:
            - FortiOS or FortiGate password.
        default: ""
    vdom:
        description:
            - Virtual domain, among those defined previously. A vdom is a
              virtual instance of the FortiGate that can be configured and
              used as a different unit.
        default: root
    https:
        description:
            - Indicates if the requests towards FortiGate must use HTTPS
              protocol
        type: bool
        default: false
    firewall_address:
        description:
            - Configure IPv4 addresses.
        default: null
        suboptions:
            state:
                description:
                    - Indicates whether to create or remove the object
                choices:
                    - present
                    - absent
            allow-routing:
                description:
                    - Enable/disable use of this address in the static route configuration.
                choices:
                    - enable
                    - disable
            associated-interface:
                description:
                    - Network interface associated with address. Source system.interface.name system.zone.name.
            cache-ttl:
                description:
                    - Defines the minimal TTL of individual IP addresses in FQDN cache measured in seconds.
            color:
                description:
                    - Color of icon on the GUI.
            comment:
                description:
                    - Comment.
            country:
                description:
                    - IP addresses associated to a specific country.
            end-ip:
                description:
                    - Final IP address (inclusive) in the range for the address.
            epg-name:
                description:
                    - Endpoint group name.
            filter:
                description:
                    - Match criteria filter.
            fqdn:
                description:
                    - Fully Qualified Domain Name address.
            list:
                description:
                    - IP address list.
                suboptions:
                    ip:
                        description:
                            - IP.
                        required: true
            name:
                description:
                    - Address name.
                required: true
            obj-id:
                description:
                    - Object ID for NSX.
            organization:
                description:
                    - "Organization domain name (Syntax: organization/domain)."
            policy-group:
                description:
                    - Policy group name.
            sdn:
                description:
                    - SDN.
                choices:
                    - aci
                    - aws
                    - azure
                    - gcp
                    - nsx
                    - nuage
                    - oci
            sdn-tag:
                description:
                    - SDN Tag.
            start-ip:
                description:
                    - First IP address (inclusive) in the range for the address.
            subnet:
                description:
                    - IP address and subnet mask of address.
            subnet-name:
                description:
                    - Subnet name.
            tagging:
                description:
                    - Config object tagging.
                suboptions:
                    category:
                        description:
                            - Tag category. Source system.object-tagging.category.
                    name:
                        description:
                            - Tagging entry name.
                        required: true
                    tags:
                        description:
                            - Tags.
                        suboptions:
                            name:
                                description:
                                    - Tag name. Source system.object-tagging.tags.name.
                                required: true
            tenant:
                description:
                    - Tenant.
            type:
                description:
                    - Type of address.
                choices:
                    - ipmask
                    - iprange
                    - fqdn
                    - geography
                    - wildcard
                    - wildcard-fqdn
                    - dynamic
            uuid:
                description:
                    - Universally Unique Identifier (UUID; automatically assigned but can be manually reset).
            visibility:
                description:
                    - Enable/disable address visibility in the GUI.
                choices:
                    - enable
                    - disable
            wildcard:
                description:
                    - IP address and wildcard netmask.
            wildcard-fqdn:
                description:
                    - Fully Qualified Domain Name with wildcard characters.
'''

EXAMPLES = '''
- hosts: localhost
  vars:
   host: "192.168.122.40"
   username: "admin"
   password: ""
   vdom: "root"
  tasks:
  - name: Configure IPv4 addresses.
    fortios_firewall_address:
      host:  "{{  host }}"
      username: "{{ username }}"
      password: "{{ password }}"
      vdom:  "{{  vdom }}"
      firewall_address:
        state: "present"
        allow-routing: "enable"
        associated-interface: "<your_own_value> (source system.interface.name system.zone.name)"
        cache-ttl: "5"
        color: "6"
        comment: "Comment."
        country: "<your_own_value>"
        end-ip: "<your_own_value>"
        epg-name: "<your_own_value>"
        filter: "<your_own_value>"
        fqdn: "<your_own_value>"
        list:
         -
            ip: "<your_own_value>"
        name: "default_name_15"
        obj-id: "<your_own_value>"
        organization: "<your_own_value>"
        policy-group: "<your_own_value>"
        sdn: "aci"
        sdn-tag: "<your_own_value>"
        start-ip: "<your_own_value>"
        subnet: "<your_own_value>"
        subnet-name: "<your_own_value>"
        tagging:
         -
            category: "<your_own_value> (source system.object-tagging.category)"
            name: "default_name_26"
            tags:
             -
                name: "default_name_28 (source system.object-tagging.tags.name)"
        tenant: "<your_own_value>"
        type: "ipmask"
        uuid: "<your_own_value>"
        visibility: "enable"
        wildcard: "<your_own_value>"
        wildcard-fqdn: "<your_own_value>"
'''

RETURN = '''
build:
  description: Build number of the fortigate image
  returned: always
  type: str
  sample: '1547'
http_method:
  description: Last method used to provision the content into FortiGate
  returned: always
  type: str
  sample: 'PUT'
http_status:
  description: Last result given by FortiGate on last operation applied
  returned: always
  type: str
  sample: "200"
mkey:
  description: Master key (id) used in the last call to FortiGate
  returned: success
  type: str
  sample: "key1"
name:
  description: Name of the table used to fulfill the request
  returned: always
  type: str
  sample: "urlfilter"
path:
  description: Path of the table used to fulfill the request
  returned: always
  type: str
  sample: "webfilter"
revision:
  description: Internal revision number
  returned: always
  type: str
  sample: "17.0.2.10658"
serial:
  description: Serial number of the unit
  returned: always
  type: str
  sample: "FGVMEVYYQT3AB5352"
status:
  description: Indication of the operation's result
  returned: always
  type: str
  sample: "success"
vdom:
  description: Virtual domain used
  returned: always
  type: str
  sample: "root"
version:
  description: Version of the FortiGate
  returned: always
  type: str
  sample: "v5.6.3"

'''

from ansible.module_utils.basic import AnsibleModule

fos = None


def login(data):
    host = data['host']
    username = data['username']
    password = data['password']

    fos.debug('on')
    if 'https' in data and not data['https']:
        fos.https('off')
    else:
        fos.https('on')

    fos.login(host, username, password)


def filter_firewall_address_data(json):
    option_list = ['allow-routing', 'associated-interface', 'cache-ttl',
                   'color', 'comment', 'country',
                   'end-ip', 'epg-name', 'filter',
                   'fqdn', 'list', 'name',
                   'obj-id', 'organization', 'policy-group',
                   'sdn', 'sdn-tag', 'start-ip',
                   'subnet', 'subnet-name', 'tagging',
                   'tenant', 'type', 'uuid',
                   'visibility', 'wildcard', 'wildcard-fqdn']
    dictionary = {}

    for attribute in option_list:
        if attribute in json:
            dictionary[attribute] = json[attribute]

    return dictionary


def firewall_address(data, fos):
    vdom = data['vdom']
    firewall_address_data = data['firewall_address']
    filtered_data = filter_firewall_address_data(firewall_address_data)
    if firewall_address_data['state'] == "present":
        return fos.set('firewall',
                       'address',
                       data=filtered_data,
                       vdom=vdom)

    elif firewall_address_data['state'] == "absent":
        return fos.delete('firewall',
                          'address',
                          mkey=filtered_data['name'],
                          vdom=vdom)


def fortios_firewall(data, fos):
    login(data)

    methodlist = ['firewall_address']
    for method in methodlist:
        if data[method]:
            resp = eval(method)(data, fos)
            break

    fos.logout()
    return not resp['status'] == "success", resp['status'] == "success", resp


def main():
    fields = {
        "host": {"required": True, "type": "str"},
        "username": {"required": True, "type": "str"},
        "password": {"required": False, "type": "str", "no_log": True},
        "vdom": {"required": False, "type": "str", "default": "root"},
        "https": {"required": False, "type": "bool", "default": "False"},
        "firewall_address": {
            "required": False, "type": "dict",
            "options": {
                "state": {"required": True, "type": "str",
                          "choices": ["present", "absent"]},
                "allow-routing": {"required": False, "type": "str",
                                  "choices": ["enable", "disable"]},
                "associated-interface": {"required": False, "type": "str"},
                "cache-ttl": {"required": False, "type": "int"},
                "color": {"required": False, "type": "int"},
                "comment": {"required": False, "type": "str"},
                "country": {"required": False, "type": "str"},
                "end-ip": {"required": False, "type": "str"},
                "epg-name": {"required": False, "type": "str"},
                "filter": {"required": False, "type": "str"},
                "fqdn": {"required": False, "type": "str"},
                "list": {"required": False, "type": "list",
                         "options": {
                             "ip": {"required": True, "type": "str"}
                         }},
                "name": {"required": True, "type": "str"},
                "obj-id": {"required": False, "type": "str"},
                "organization": {"required": False, "type": "str"},
                "policy-group": {"required": False, "type": "str"},
                "sdn": {"required": False, "type": "str",
                        "choices": ["aci", "aws", "azure",
                                    "gcp", "nsx", "nuage",
                                    "oci"]},
                "sdn-tag": {"required": False, "type": "str"},
                "start-ip": {"required": False, "type": "str"},
                "subnet": {"required": False, "type": "str"},
                "subnet-name": {"required": False, "type": "str"},
                "tagging": {"required": False, "type": "list",
                            "options": {
                                "category": {"required": False, "type": "str"},
                                "name": {"required": True, "type": "str"},
                                "tags": {"required": False, "type": "list",
                                         "options": {
                                             "name": {"required": True, "type": "str"}
                                         }}
                            }},
                "tenant": {"required": False, "type": "str"},
                "type": {"required": False, "type": "str",
                         "choices": ["ipmask", "iprange", "fqdn",
                                     "geography", "wildcard", "wildcard-fqdn",
                                     "dynamic"]},
                "uuid": {"required": False, "type": "str"},
                "visibility": {"required": False, "type": "str",
                               "choices": ["enable", "disable"]},
                "wildcard": {"required": False, "type": "str"},
                "wildcard-fqdn": {"required": False, "type": "str"}

            }
        }
    }

    module = AnsibleModule(argument_spec=fields,
                           supports_check_mode=False)
    try:
        from fortiosapi import FortiOSAPI
    except ImportError:
        module.fail_json(msg="fortiosapi module is required")

    global fos
    fos = FortiOSAPI()

    is_error, has_changed, result = fortios_firewall(module.params, fos)

    if not is_error:
        module.exit_json(changed=has_changed, meta=result)
    else:
        module.fail_json(msg="Error in repo", meta=result)


if __name__ == '__main__':
    main()



# Copyright (c) 2018, Oracle and/or its affiliates.
# This software is made available to you under the terms of the GPL 3.0 license or the Apache 2.0 license.
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# Apache License v2.0
# See LICENSE.TXT for details.


class ModuleDocFragment(object):
    DOCUMENTATION = """
    options:
        display_name:
            description: Use I(display_name) along with the other options to return only resources that match the given
                         display name exactly.
            type: str
    """



# (c) 2018-2019, NetApp, Inc
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

''' unit tests for ONTAP Ansible module na_ontap_gather_facts '''

from __future__ import print_function
import json
import pytest
import sys

from units.compat import unittest
from units.compat.mock import patch
from ansible.module_utils import basic
from ansible.module_utils._text import to_bytes
import ansible.module_utils.netapp as netapp_utils

from ansible.modules.storage.netapp.na_ontap_gather_facts import main as gather_facts_main
from ansible.modules.storage.netapp.na_ontap_gather_facts import __finditem as gather_facts_finditem
from ansible.modules.storage.netapp.na_ontap_gather_facts \
    import NetAppONTAPGatherFacts as gather_facts_module  # module under test

if not netapp_utils.has_netapp_lib():
    pytestmark = pytest.mark.skip('skipping as missing required netapp_lib')


def set_module_args(args):
    """prepare arguments so that they will be picked up during module creation"""
    args = json.dumps({'ANSIBLE_MODULE_ARGS': args})
    basic._ANSIBLE_ARGS = to_bytes(args)  # pylint: disable=protected-access


class AnsibleExitJson(Exception):
    """Exception class to be raised by module.exit_json and caught by the test case"""
    pass


class AnsibleFailJson(Exception):
    """Exception class to be raised by module.fail_json and caught by the test case"""
    pass


def exit_json(*args, **kwargs):  # pylint: disable=unused-argument
    """function to patch over exit_json; package return data into an exception"""
    if 'changed' not in kwargs:
        kwargs['changed'] = False
    raise AnsibleExitJson(kwargs)


def fail_json(*args, **kwargs):  # pylint: disable=unused-argument
    """function to patch over fail_json; package return data into an exception"""
    kwargs['failed'] = True
    raise AnsibleFailJson(kwargs)


class MockONTAPConnection(object):
    ''' mock server connection to ONTAP host '''

    def __init__(self, kind=None):
        ''' save arguments '''
        self.type = kind
        self.xml_in = None
        self.xml_out = None

    def invoke_successfully(self, xml, enable_tunneling):  # pylint: disable=unused-argument
        ''' mock invoke_successfully returning xml data '''
        self.xml_in = xml
        print(xml.to_string())
        if self.type == 'vserver':
            xml = self.build_vserver_info()
        elif self.type == 'net_port':
            xml = self.build_net_port_info()
        elif self.type == 'zapi_error':
            error = netapp_utils.zapi.NaApiError('test', 'error')
            raise error
        self.xml_out = xml
        return xml

    @staticmethod
    def build_vserver_info():
        ''' build xml data for vserser-info '''
        xml = netapp_utils.zapi.NaElement('xml')
        attributes = netapp_utils.zapi.NaElement('attributes-list')
        attributes.add_node_with_children('vserver-info',
                                          **{'vserver-name': 'test_vserver'})
        xml.add_child_elem(attributes)
        return xml

    @staticmethod
    def build_net_port_info():
        ''' build xml data for net-port-info '''
        xml = netapp_utils.zapi.NaElement('xml')
        attributes_list = netapp_utils.zapi.NaElement('attributes-list')
        num_net_port_info = 2
        for i in range(num_net_port_info):
            net_port_info = netapp_utils.zapi.NaElement('net-port-info')
            net_port_info.add_new_child('node', 'node_' + str(i))
            net_port_info.add_new_child('port', 'port_' + str(i))
            net_port_info.add_new_child('broadcast_domain', 'test_domain_' + str(i))
            net_port_info.add_new_child('ipspace', 'ipspace' + str(i))
            attributes_list.add_child_elem(net_port_info)
        xml.add_child_elem(attributes_list)
        return xml


class TestMyModule(unittest.TestCase):
    ''' a group of related Unit Tests '''

    def setUp(self):
        self.mock_module_helper = patch.multiple(basic.AnsibleModule,
                                                 exit_json=exit_json,
                                                 fail_json=fail_json)
        self.mock_module_helper.start()
        self.addCleanup(self.mock_module_helper.stop)
        self.server = MockONTAPConnection()

    def mock_args(self):
        return {
            'hostname': 'hostname',
            'username': 'username',
            'password': 'password',
        }

    def get_gather_facts_mock_object(self, kind=None):
        """
        Helper method to return an na_ontap_gather_facts object
        """
        module = basic.AnsibleModule(
            argument_spec=netapp_utils.na_ontap_host_argument_spec(),
            supports_check_mode=True
        )
        obj = gather_facts_module(module)
        obj.netapp_info = dict()
        if kind is None:
            obj.server = MockONTAPConnection()
        else:
            obj.server = MockONTAPConnection(kind)
        return obj

    def test_module_fail_when_required_args_missing(self):
        ''' required arguments are reported as errors '''
        with pytest.raises(AnsibleFailJson) as exc:
            set_module_args({})
            self.get_gather_facts_mock_object()
        print('Info: %s' % exc.value.args[0]['msg'])

    @patch('ansible.module_utils.netapp.ems_log_event')
    def test_ensure_command_called(self, mock_ems_log):
        ''' calling get_all will raise a KeyError exception '''
        set_module_args(self.mock_args())
        my_obj = self.get_gather_facts_mock_object('vserver')
        with pytest.raises(KeyError) as exc:
            my_obj.get_all(['net_interface_info'])
        if sys.version_info >= (2, 7):
            msg = 'net-interface-info'
            assert exc.value.args[0] == msg

    @patch('ansible.module_utils.netapp.ems_log_event')
    def test_get_generic_get_iter(self, mock_ems_log):
        '''calling get_generic_get_iter will return expected dict'''
        set_module_args(self.mock_args())
        obj = self.get_gather_facts_mock_object('net_port')
        result = obj.get_generic_get_iter(
            'net-port-get-iter',
            attribute='net-port-info',
            field=('node', 'port'),
            query={'max-records': '1024'}
        )
        assert result.get('node_0:port_0')
        assert result.get('node_1:port_1')

    @patch('ansible.modules.storage.netapp.na_ontap_gather_facts.NetAppONTAPGatherFacts.get_all')
    def test_main(self, get_all):
        '''test main method.'''
        set_module_args(self.mock_args())
        get_all.side_effect = [
            {'test_get_all':
                {'vserver_login_banner_info': 'test_vserver_login_banner_info', 'vserver_info': 'test_vserver_info'}}
        ]
        with pytest.raises(AnsibleExitJson) as exc:
            gather_facts_main()
        assert exc.value.args[0]['state'] == 'info'

    @patch('ansible.modules.storage.netapp.na_ontap_gather_facts.NetAppONTAPGatherFacts.get_generic_get_iter')
    def test_get_ifgrp_info(self, get_generic_get_iter):
        '''test get_ifgrp_info with empty ifgrp_info'''
        set_module_args(self.mock_args())
        get_generic_get_iter.side_effect = [
            {}
        ]
        obj = self.get_gather_facts_mock_object()
        obj.netapp_info['net_port_info'] = {}
        result = obj.get_ifgrp_info()
        assert result == {}

    def test_ontapi_error(self):
        '''test ontapi will raise zapi error'''
        set_module_args(self.mock_args())
        obj = self.get_gather_facts_mock_object('zapi_error')
        with pytest.raises(AnsibleFailJson) as exc:
            obj.ontapi()
        assert exc.value.args[0]['msg'] == 'Error calling API system-get-ontapi-version: NetApp API failed. Reason - test:error'

    def test_call_api_error(self):
        '''test call_api will raise zapi error'''
        set_module_args(self.mock_args())
        obj = self.get_gather_facts_mock_object('zapi_error')
        with pytest.raises(AnsibleFailJson) as exc:
            obj.call_api('nvme-get-iter')
        assert exc.value.args[0]['msg'] == 'Error calling API nvme-get-iter: NetApp API failed. Reason - test:error'

    def test_find_item(self):
        '''test __find_item return expected key value'''
        obj = {"A": 1, "B": {"C": {"D": 2}}}
        key = "D"
        result = gather_facts_finditem(obj, key)
        assert result == 2

    def test_subset_return_all_complete(self):
        ''' Check all returns all of the entries if version is high enough '''
        version = '140'         # change this if new ZAPIs are supported
        set_module_args(self.mock_args())
        obj = self.get_gather_facts_mock_object('vserver')
        subset = obj.get_subset(['all'], version)
        assert set(obj.fact_subsets.keys()) == subset

    def test_subset_return_all_partial(self):
        ''' Check all returns a subset of the entries if version is low enough '''
        version = '120'         # low enough so that some ZAPIs are not supported
        set_module_args(self.mock_args())
        obj = self.get_gather_facts_mock_object('vserver')
        subset = obj.get_subset(['all'], version)
        all_keys = obj.fact_subsets.keys()
        assert set(all_keys) > subset
        supported_keys = filter(lambda key: obj.fact_subsets[key]['min_version'] <= version, all_keys)
        assert set(supported_keys) == subset

    def test_subset_return_one(self):
        ''' Check single entry returns one '''
        version = '120'         # low enough so that some ZAPIs are not supported
        set_module_args(self.mock_args())
        obj = self.get_gather_facts_mock_object('vserver')
        subset = obj.get_subset(['net_interface_info'], version)
        assert len(subset) == 1

    def test_subset_return_multiple(self):
        ''' Check that more than one entry returns the same number '''
        version = '120'         # low enough so that some ZAPIs are not supported
        set_module_args(self.mock_args())
        obj = self.get_gather_facts_mock_object('vserver')
        subset_entries = ['net_interface_info', 'net_port_info']
        subset = obj.get_subset(subset_entries, version)
        assert len(subset) == len(subset_entries)

    def test_subset_return_bad(self):
        ''' Check that a bad subset entry will error out '''
        version = '120'         # low enough so that some ZAPIs are not supported
        set_module_args(self.mock_args())
        obj = self.get_gather_facts_mock_object('vserver')
        with pytest.raises(AnsibleFailJson) as exc:
            subset = obj.get_subset(['net_interface_info', 'my_invalid_subset'], version)
        print('Info: %s' % exc.value.args[0]['msg'])
        assert exc.value.args[0]['msg'] == 'Bad subset: my_invalid_subset'

    def test_subset_return_unsupported(self):
        ''' Check that a new subset entry will error out on an older system '''
        version = '120'         # low enough so that some ZAPIs are not supported
        key = 'nvme_info'       # only supported starting at 140
        set_module_args(self.mock_args())
        obj = self.get_gather_facts_mock_object('vserver')
        with pytest.raises(AnsibleFailJson) as exc:
            subset = obj.get_subset(['net_interface_info', key], version)
        print('Info: %s' % exc.value.args[0]['msg'])
        msg = 'Remote system at version %s does not support %s' % (version, key)
        assert exc.value.args[0]['msg'] == msg

    def test_subset_return_none(self):
        ''' Check usable subset can be empty '''
        version = '!'   # lower then 0, so that no ZAPI is supported
        set_module_args(self.mock_args())
        obj = self.get_gather_facts_mock_object('vserver')
        subset = obj.get_subset(['all'], version)
        assert len(subset) == 0

    def test_subset_return_all_expect_one(self):
        ''' Check !x returns all of the entries except x if version is high enough '''
        version = '140'         # change this if new ZAPIs are supported
        set_module_args(self.mock_args())
        obj = self.get_gather_facts_mock_object('vserver')
        subset = obj.get_subset(['!net_interface_info'], version)
        assert len(obj.fact_subsets.keys()) == len(subset) + 1
        subset.add('net_interface_info')
        assert set(obj.fact_subsets.keys()) == subset

    def test_subset_return_all_expect_three(self):
        ''' Check !x,!y,!z returns all of the entries except x, y, z if version is high enough '''
        version = '140'         # change this if new ZAPIs are supported
        set_module_args(self.mock_args())
        obj = self.get_gather_facts_mock_object('vserver')
        subset = obj.get_subset(['!net_interface_info', '!nvme_info', '!ontap_version'], version)
        assert len(obj.fact_subsets.keys()) == len(subset) + 3
        subset.update(['net_interface_info', 'nvme_info', 'ontap_version'])
        assert set(obj.fact_subsets.keys()) == subset

    def test_subset_return_none_with_exclusion(self):
        ''' Check usable subset can be empty with !x '''
        version = '!'   # lower then 0, so that no ZAPI is supported
        key = 'net_interface_info'
        set_module_args(self.mock_args())
        obj = self.get_gather_facts_mock_object('vserver')
        with pytest.raises(AnsibleFailJson) as exc:
            subset = obj.get_subset(['!' + key], version)
        print('Info: %s' % exc.value.args[0]['msg'])
        msg = 'Remote system at version %s does not support %s' % (version, key)
        assert exc.value.args[0]['msg'] == msg



# (c) 2016 Red Hat Inc.
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.

# Make coding more python3-ish
from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

from units.compat.mock import patch, MagicMock
from ansible.modules.network.cli import cli_config
from units.modules.utils import set_module_args
from .cli_module import TestCliModule, load_fixture


class TestCliConfigModule(TestCliModule):

    module = cli_config

    def setUp(self):
        super(TestCliConfigModule, self).setUp()

        self.mock_connection = patch('ansible.modules.network.cli.cli_config.Connection')
        self.get_connection = self.mock_connection.start()

        self.conn = self.get_connection()

    def tearDown(self):
        super(TestCliConfigModule, self).tearDown()

        self.mock_connection.stop()

    @patch('ansible.modules.network.cli.cli_config.run')
    def test_cli_config_backup_returns__backup__(self, run_mock):
        self.conn.get_capabilities = MagicMock(return_value='{}')

        args = dict(backup=True)
        set_module_args(args)

        run_mock.return_value = {}

        result = self.execute_module()
        self.assertIn('__backup__', result)



# coding=utf-8
# (c) 2018, NetApp Inc.
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

try:
    from unittest import mock
except ImportError:
    import mock

from ansible.module_utils.netapp import NetAppESeriesModule
from ansible.modules.storage.netapp.netapp_e_volume import NetAppESeriesVolume
from units.modules.utils import AnsibleExitJson, AnsibleFailJson, ModuleTestCase, set_module_args

__metaclass__ = type


class NetAppESeriesVolumeTest(ModuleTestCase):
    REQUIRED_PARAMS = {"api_username": "username",
                       "api_password": "password",
                       "api_url": "http://localhost/devmgr/v2",
                       "ssid": "1",
                       "validate_certs": "no"}

    THIN_VOLUME_RESPONSE = [{"capacity": "1288490188800",
                             "volumeRef": "3A000000600A098000A4B28D000010475C405428",
                             "status": "optimal",
                             "protectionType": "type1Protection",
                             "maxVirtualCapacity": "281474976710656",
                             "initialProvisionedCapacity": "4294967296",
                             "currentProvisionedCapacity": "4294967296",
                             "provisionedCapacityQuota": "1305670057984",
                             "growthAlertThreshold": 85,
                             "expansionPolicy": "automatic",
                             "flashCached": False,
                             "metadata": [{"key": "workloadId", "value": "4200000001000000000000000000000000000000"},
                                          {"key": "volumeTypeId", "value": "volume"}],
                             "dataAssurance": True,
                             "segmentSize": 131072,
                             "diskPool": True,
                             "listOfMappings": [],
                             "mapped": False,
                             "currentControllerId": "070000000000000000000001",
                             "cacheSettings": {"readCacheEnable": True, "writeCacheEnable": True,
                                               "readAheadMultiplier": 0},
                             "name": "thin_volume",
                             "id": "3A000000600A098000A4B28D000010475C405428"}]
    VOLUME_GET_RESPONSE = [{"offline": False,
                            "raidLevel": "raid6",
                            "capacity": "214748364800",
                            "reconPriority": 1,
                            "segmentSize": 131072,
                            "volumeRef": "02000000600A098000A4B9D100000F095C2F7F31",
                            "status": "optimal",
                            "protectionInformationCapable": False,
                            "protectionType": "type0Protection",
                            "diskPool": True,
                            "flashCached": False,
                            "metadata": [{"key": "workloadId", "value": "4200000002000000000000000000000000000000"},
                                         {"key": "volumeTypeId", "value": "Clare"}],
                            "dataAssurance": False,
                            "currentControllerId": "070000000000000000000002",
                            "cacheSettings": {"readCacheEnable": True, "writeCacheEnable": False,
                                              "readAheadMultiplier": 0},
                            "thinProvisioned": False,
                            "totalSizeInBytes": "214748364800",
                            "name": "Matthew",
                            "id": "02000000600A098000A4B9D100000F095C2F7F31"},
                           {"offline": False,
                            "raidLevel": "raid6",
                            "capacity": "107374182400",
                            "reconPriority": 1,
                            "segmentSize": 131072,
                            "volumeRef": "02000000600A098000A4B28D00000FBE5C2F7F26",
                            "status": "optimal",
                            "protectionInformationCapable": False,
                            "protectionType": "type0Protection",
                            "diskPool": True,
                            "flashCached": False,
                            "metadata": [{"key": "workloadId", "value": "4200000002000000000000000000000000000000"},
                                         {"key": "volumeTypeId", "value": "Samantha"}],
                            "dataAssurance": False,
                            "currentControllerId": "070000000000000000000001",
                            "cacheSettings": {"readCacheEnable": True, "writeCacheEnable": False,
                                              "readAheadMultiplier": 0},
                            "thinProvisioned": False,
                            "totalSizeInBytes": "107374182400",
                            "name": "Samantha",
                            "id": "02000000600A098000A4B28D00000FBE5C2F7F26"},
                           {"offline": False,
                            "raidLevel": "raid6",
                            "capacity": "107374182400",
                            "segmentSize": 131072,
                            "volumeRef": "02000000600A098000A4B9D100000F0B5C2F7F40",
                            "status": "optimal",
                            "protectionInformationCapable": False,
                            "protectionType": "type0Protection",
                            "volumeGroupRef": "04000000600A098000A4B9D100000F085C2F7F26",
                            "diskPool": True,
                            "flashCached": False,
                            "metadata": [{"key": "workloadId", "value": "4200000002000000000000000000000000000000"},
                                         {"key": "volumeTypeId", "value": "Micah"}],
                            "dataAssurance": False,
                            "currentControllerId": "070000000000000000000002",
                            "cacheSettings": {"readCacheEnable": True, "writeCacheEnable": False,
                                              "readAheadMultiplier": 0},
                            "thinProvisioned": False,
                            "totalSizeInBytes": "107374182400",
                            "name": "Micah",
                            "id": "02000000600A098000A4B9D100000F0B5C2F7F40"}]
    STORAGE_POOL_GET_RESPONSE = [{"offline": False,
                                  "raidLevel": "raidDiskPool",
                                  "volumeGroupRef": "04000000600A",
                                  "securityType": "capable",
                                  "protectionInformationCapable": False,
                                  "protectionInformationCapabilities": {"protectionInformationCapable": True,
                                                                        "protectionType": "type2Protection"},
                                  "volumeGroupData": {"type": "diskPool",
                                                      "diskPoolData": {"reconstructionReservedDriveCount": 1,
                                                                       "reconstructionReservedAmt": "296889614336",
                                                                       "reconstructionReservedDriveCountCurrent": 1,
                                                                       "poolUtilizationWarningThreshold": 0,
                                                                       "poolUtilizationCriticalThreshold": 85,
                                                                       "poolUtilizationState": "utilizationOptimal",
                                                                       "unusableCapacity": "0",
                                                                       "degradedReconstructPriority": "high",
                                                                       "criticalReconstructPriority": "highest",
                                                                       "backgroundOperationPriority": "low",
                                                                       "allocGranularity": "4294967296"}},
                                  "reservedSpaceAllocated": False,
                                  "securityLevel": "fde",
                                  "usedSpace": "863288426496",
                                  "totalRaidedSpace": "2276332666880",
                                  "raidStatus": "optimal",
                                  "freeSpace": "1413044240384",
                                  "drivePhysicalType": "sas",
                                  "driveMediaType": "hdd",
                                  "diskPool": True,
                                  "id": "04000000600A098000A4B9D100000F085C2F7F26",
                                  "name": "employee_data_storage_pool"},
                                 {"offline": False,
                                  "raidLevel": "raid1",
                                  "volumeGroupRef": "04000000600A098000A4B28D00000FBD5C2F7F19",
                                  "state": "complete",
                                  "securityType": "capable",
                                  "drawerLossProtection": False,
                                  "protectionInformationCapable": False,
                                  "protectionInformationCapabilities": {"protectionInformationCapable": True,
                                                                        "protectionType": "type2Protection"},
                                  "volumeGroupData": {"type": "unknown", "diskPoolData": None},
                                  "reservedSpaceAllocated": False,
                                  "securityLevel": "fde",
                                  "usedSpace": "322122547200",
                                  "totalRaidedSpace": "598926258176",
                                  "raidStatus": "optimal",
                                  "freeSpace": "276803710976",
                                  "drivePhysicalType": "sas",
                                  "driveMediaType": "hdd",
                                  "diskPool": False,
                                  "id": "04000000600A098000A4B28D00000FBD5C2F7F19",
                                  "name": "database_storage_pool"}]
    REQUEST_FUNC = "ansible.modules.storage.netapp.netapp_e_volume.NetAppESeriesVolume.request"

    WORKLOAD_GET_RESPONSE = [{"id": "4200000001000000000000000000000000000000", "name": "general_workload_1",
                              "workloadAttributes": [{"key": "profileId", "value": "Other_1"}]},
                             {"id": "4200000002000000000000000000000000000000", "name": "employee_data",
                              "workloadAttributes": [{"key": "use", "value": "EmployeeData"},
                                                     {"key": "location", "value": "ICT"},
                                                     {"key": "private", "value": "public"},
                                                     {"key": "profileId", "value": "ansible_workload_1"}]},
                             {"id": "4200000003000000000000000000000000000000", "name": "customer_database",
                              "workloadAttributes": [{"key": "use", "value": "customer_information"},
                                                     {"key": "location", "value": "global"},
                                                     {"key": "profileId", "value": "ansible_workload_2"}]},
                             {"id": "4200000004000000000000000000000000000000", "name": "product_database",
                              "workloadAttributes": [{"key": "use", "value": "production_information"},
                                                     {"key": "security", "value": "private"},
                                                     {"key": "location", "value": "global"},
                                                     {"key": "profileId", "value": "ansible_workload_4"}]}]

    def _set_args(self, args=None):
        module_args = self.REQUIRED_PARAMS.copy()
        if args is not None:
            module_args.update(args)
        set_module_args(module_args)

    def test_module_arguments_pass(self):
        """Ensure valid arguments successful create a class instance."""
        arg_sets = [{"state": "present", "name": "vol", "storage_pool_name": "pool", "size": 100, "size_unit": "tb",
                     "thin_provision": True, "thin_volume_repo_size": 64, "thin_volume_max_repo_size": 1000,
                     "thin_volume_growth_alert_threshold": 10},
                    {"state": "present", "name": "vol", "storage_pool_name": "pool", "size": 100, "size_unit": "gb",
                     "thin_provision": True, "thin_volume_repo_size": 64, "thin_volume_max_repo_size": 1024,
                     "thin_volume_growth_alert_threshold": 99},
                    {"state": "present", "name": "vol", "storage_pool_name": "pool", "size": 100, "size_unit": "gb",
                     "thin_provision": True, "thin_volume_repo_size": 64},
                    {"state": "present", "name": "vol", "storage_pool_name": "pool", "size": 100, "size_unit": "kb",
                     "thin_provision": True, "thin_volume_repo_size": 64, "thin_volume_max_repo_size": 67108864}]

        # validate size normalization
        for arg_set in arg_sets:
            self._set_args(arg_set)
            volume_object = NetAppESeriesVolume()

            size_unit_multiplier = NetAppESeriesModule.SIZE_UNIT_MAP[arg_set["size_unit"]]
            self.assertEqual(volume_object.size_b, arg_set["size"] * size_unit_multiplier)
            self.assertEqual(volume_object.thin_volume_repo_size_b,
                             arg_set["thin_volume_repo_size"] * size_unit_multiplier)
            self.assertEqual(volume_object.thin_volume_expansion_policy, "automatic")
            if "thin_volume_max_repo_size" not in arg_set.keys():
                self.assertEqual(volume_object.thin_volume_max_repo_size_b, arg_set["size"] * size_unit_multiplier)
            else:
                self.assertEqual(volume_object.thin_volume_max_repo_size_b,
                                 arg_set["thin_volume_max_repo_size"] * size_unit_multiplier)

        # validate metadata form
        self._set_args(
            {"state": "present", "name": "vol", "storage_pool_name": "pool", "size": 10, "workload_name": "workload1",
             "metadata": {"availability": "public", "security": "low"}})
        volume_object = NetAppESeriesVolume()
        for entry in volume_object.metadata:
            self.assertTrue(entry in [{'value': 'low', 'key': 'security'}, {'value': 'public', 'key': 'availability'}])

    def test_module_arguments_fail(self):
        """Ensure invalid arguments values do not create a class instance."""
        arg_sets = [{"state": "present", "name": "vol", "storage_pool_name": "pool", "size": 100, "size_unit": "tb",
                     "thin_provision": True, "thin_volume_repo_size": 260},
                    {"state": "present", "name": "vol", "storage_pool_name": "pool", "size": 10000, "size_unit": "tb",
                     "thin_provision": True, "thin_volume_repo_size": 64, "thin_volume_max_repo_size": 10},
                    {"state": "present", "name": "vol", "storage_pool_name": "pool", "size": 10000, "size_unit": "gb",
                     "thin_provision": True, "thin_volume_repo_size": 64, "thin_volume_max_repo_size": 1000,
                     "thin_volume_growth_alert_threshold": 9},
                    {"state": "present", "name": "vol", "storage_pool_name": "pool", "size": 10000, "size_unit": "gb",
                     "thin_provision": True, "thin_volume_repo_size": 64, "thin_volume_max_repo_size": 1000,
                     "thin_volume_growth_alert_threshold": 100}]

        for arg_set in arg_sets:
            with self.assertRaises(AnsibleFailJson):
                self._set_args(arg_set)
                print(arg_set)
                volume_object = NetAppESeriesVolume()

    def test_get_volume_pass(self):
        """Evaluate the get_volume method."""
        with mock.patch(self.REQUEST_FUNC,
                        side_effect=[(200, self.VOLUME_GET_RESPONSE), (200, self.THIN_VOLUME_RESPONSE)]):
            self._set_args({"state": "present", "name": "Matthew", "storage_pool_name": "pool", "size": 100})
            volume_object = NetAppESeriesVolume()
            self.assertEqual(volume_object.get_volume(),
                             [entry for entry in self.VOLUME_GET_RESPONSE if entry["name"] == "Matthew"][0])

        with mock.patch(self.REQUEST_FUNC,
                        side_effect=[(200, self.VOLUME_GET_RESPONSE), (200, self.THIN_VOLUME_RESPONSE)]):
            self._set_args({"state": "present", "name": "NotAVolume", "storage_pool_name": "pool", "size": 100})
            volume_object = NetAppESeriesVolume()
            self.assertEqual(volume_object.get_volume(), {})

    def test_get_volume_fail(self):
        """Evaluate the get_volume exception paths."""
        with self.assertRaisesRegexp(AnsibleFailJson, "Failed to obtain list of thick volumes."):
            with mock.patch(self.REQUEST_FUNC, return_value=Exception()):
                self._set_args({"state": "present", "name": "Matthew", "storage_pool_name": "pool", "size": 100})
                volume_object = NetAppESeriesVolume()
                volume_object.get_volume()

        with self.assertRaisesRegexp(AnsibleFailJson, "Failed to obtain list of thin volumes."):
            with mock.patch(self.REQUEST_FUNC, side_effect=[(200, self.VOLUME_GET_RESPONSE), Exception()]):
                self._set_args({"state": "present", "name": "Matthew", "storage_pool_name": "pool", "size": 100})
                volume_object = NetAppESeriesVolume()
                volume_object.get_volume()

    def test_get_storage_pool_pass(self):
        """Evaluate the get_storage_pool method."""
        with mock.patch(self.REQUEST_FUNC, return_value=(200, self.STORAGE_POOL_GET_RESPONSE)):
            self._set_args({"state": "present", "name": "NewVolume", "storage_pool_name": "employee_data_storage_pool",
                            "size": 100})
            volume_object = NetAppESeriesVolume()
            self.assertEqual(volume_object.get_storage_pool(), [entry for entry in self.STORAGE_POOL_GET_RESPONSE if
                                                                entry["name"] == "employee_data_storage_pool"][0])

            self._set_args(
                {"state": "present", "name": "NewVolume", "storage_pool_name": "NotAStoragePool", "size": 100})
            volume_object = NetAppESeriesVolume()
            self.assertEqual(volume_object.get_storage_pool(), {})

    def test_get_storage_pool_fail(self):
        """Evaluate the get_storage_pool exception paths."""
        with self.assertRaisesRegexp(AnsibleFailJson, "Failed to obtain list of storage pools."):
            with mock.patch(self.REQUEST_FUNC, return_value=Exception()):
                self._set_args({"state": "present", "name": "Matthew", "storage_pool_name": "pool", "size": 100})
                volume_object = NetAppESeriesVolume()
                volume_object.get_storage_pool()

    def test_check_storage_pool_sufficiency_pass(self):
        """Ensure passing logic."""
        self._set_args({"state": "present", "name": "Matthew", "storage_pool_name": "pool", "size": 100})
        volume_object = NetAppESeriesVolume()
        volume_object.pool_detail = [entry for entry in self.STORAGE_POOL_GET_RESPONSE
                                     if entry["name"] == "employee_data_storage_pool"][0]
        volume_object.check_storage_pool_sufficiency()

    def test_check_storage_pool_sufficiency_fail(self):
        """Validate exceptions are thrown for insufficient storage pool resources."""
        self._set_args({"state": "present", "name": "vol", "storage_pool_name": "pool", "size": 100, "size_unit": "tb",
                        "thin_provision": True, "thin_volume_repo_size": 64, "thin_volume_max_repo_size": 1000,
                        "thin_volume_growth_alert_threshold": 10})
        volume_object = NetAppESeriesVolume()

        with self.assertRaisesRegexp(AnsibleFailJson, "Requested storage pool"):
            volume_object.check_storage_pool_sufficiency()

        with self.assertRaisesRegexp(AnsibleFailJson,
                                     "Thin provisioned volumes can only be created on raid disk pools."):
            volume_object.pool_detail = [entry for entry in self.STORAGE_POOL_GET_RESPONSE
                                         if entry["name"] == "database_storage_pool"][0]
            volume_object.volume_detail = {}
            volume_object.check_storage_pool_sufficiency()

        with self.assertRaisesRegexp(AnsibleFailJson, "requires the storage pool to be DA-compatible."):
            volume_object.pool_detail = {"diskPool": True,
                                         "protectionInformationCapabilities": {"protectionType": "type0Protection",
                                                                               "protectionInformationCapable": False}}
            volume_object.volume_detail = {}
            volume_object.data_assurance_enabled = True
            volume_object.check_storage_pool_sufficiency()

            volume_object.pool_detail = {"diskPool": True,
                                         "protectionInformationCapabilities": {"protectionType": "type2Protection",
                                                                               "protectionInformationCapable": True}}
            volume_object.check_storage_pool_sufficiency()

        self._set_args({"state": "present", "name": "vol", "storage_pool_name": "pool", "size": 100, "size_unit": "tb",
                        "thin_provision": False})
        volume_object = NetAppESeriesVolume()
        with self.assertRaisesRegexp(AnsibleFailJson,
                                     "Not enough storage pool free space available for the volume's needs."):
            volume_object.pool_detail = {"freeSpace": 10, "diskPool": True,
                                         "protectionInformationCapabilities": {"protectionType": "type2Protection",
                                                                               "protectionInformationCapable": True}}
            volume_object.volume_detail = {"totalSizeInBytes": 100}
            volume_object.data_assurance_enabled = True
            volume_object.size_b = 1
            volume_object.check_storage_pool_sufficiency()

    def test_update_workload_tags_pass(self):
        """Validate updating workload tags."""
        test_sets = [[{"state": "present", "name": "Matthew", "storage_pool_name": "pool", "size": 100}, False],
                     [{"state": "present", "name": "Matthew", "storage_pool_name": "pool", "size": 100,
                       "workload_name": "employee_data"}, False],
                     [{"state": "present", "name": "Matthew", "storage_pool_name": "pool", "size": 100,
                       "workload_name": "customer_database",
                       "metadata": {"use": "customer_information", "location": "global"}}, False],
                     [{"state": "present", "name": "Matthew", "storage_pool_name": "pool", "size": 100,
                       "workload_name": "customer_database",
                       "metadata": {"use": "customer_information"}}, True],
                     [{"state": "present", "name": "Matthew", "storage_pool_name": "pool", "size": 100,
                       "workload_name": "customer_database",
                       "metadata": {"use": "customer_information", "location": "local"}}, True],
                     [{"state": "present", "name": "Matthew", "storage_pool_name": "pool", "size": 100,
                       "workload_name": "customer_database",
                       "metadata": {"use": "customer_information", "location": "global", "importance": "no"}}, True],
                     [{"state": "present", "name": "Matthew", "storage_pool_name": "pool", "size": 100,
                       "workload_name": "newWorkload",
                       "metadata": {"for_testing": "yes"}}, True],
                     [{"state": "present", "name": "Matthew", "storage_pool_name": "pool", "size": 100,
                       "workload_name": "newWorkload"}, True]]

        for test in test_sets:
            self._set_args(test[0])
            volume_object = NetAppESeriesVolume()

            with mock.patch(self.REQUEST_FUNC, side_effect=[(200, self.WORKLOAD_GET_RESPONSE), (200, {"id": 1})]):
                self.assertEqual(volume_object.update_workload_tags(), test[1])

    def test_update_workload_tags_fail(self):
        """Validate updating workload tags fails appropriately."""
        self._set_args({"state": "present", "name": "Matthew", "storage_pool_name": "pool", "size": 100,
                        "workload_name": "employee_data"})
        volume_object = NetAppESeriesVolume()
        with self.assertRaisesRegexp(AnsibleFailJson, "Failed to retrieve storage array workload tags."):
            with mock.patch(self.REQUEST_FUNC, return_value=Exception()):
                volume_object.update_workload_tags()

        self._set_args({"state": "present", "name": "Matthew", "storage_pool_name": "pool", "size": 100,
                        "workload_name": "employee_data", "metadata": {"key": "not-use", "value": "EmployeeData"}})
        volume_object = NetAppESeriesVolume()
        with self.assertRaisesRegexp(AnsibleFailJson, "Failed to create new workload tag."):
            with mock.patch(self.REQUEST_FUNC, side_effect=[(200, self.WORKLOAD_GET_RESPONSE), Exception()]):
                volume_object.update_workload_tags()

        self._set_args({"state": "present", "name": "Matthew", "storage_pool_name": "pool", "size": 100,
                        "workload_name": "employee_data2", "metadata": {"key": "use", "value": "EmployeeData"}})
        volume_object = NetAppESeriesVolume()
        with self.assertRaisesRegexp(AnsibleFailJson, "Failed to create new workload tag."):
            with mock.patch(self.REQUEST_FUNC, side_effect=[(200, self.WORKLOAD_GET_RESPONSE), Exception()]):
                volume_object.update_workload_tags()

    def test_get_volume_property_changes_pass(self):
        """Verify correct dictionary is returned"""

        # no property changes
        self._set_args(
            {"state": "present", "name": "Matthew", "storage_pool_name": "pool", "size": 100, "ssd_cache_enabled": True,
             "read_cache_enable": True, "write_cache_enable": True,
             "read_ahead_enable": True, "thin_provision": False})
        volume_object = NetAppESeriesVolume()
        volume_object.volume_detail = {"metadata": [],
                                       "cacheSettings": {"readCacheEnable": True, "writeCacheEnable": True,
                                                         "readAheadMultiplier": 1}, "flashCached": True,
                                       "segmentSize": str(128 * 1024)}
        self.assertEqual(volume_object.get_volume_property_changes(), dict())

        self._set_args(
            {"state": "present", "name": "Matthew", "storage_pool_name": "pool", "size": 100, "ssd_cache_enabled": True,
             "read_cache_enable": True, "write_cache_enable": True,
             "read_ahead_enable": True, "thin_provision": True, "thin_volume_repo_size": 64,
             "thin_volume_max_repo_size": 1000, "thin_volume_growth_alert_threshold": 90})
        volume_object = NetAppESeriesVolume()
        volume_object.volume_detail = {"metadata": [],
                                       "cacheSettings": {"readCacheEnable": True, "writeCacheEnable": True,
                                                         "readAheadMultiplier": 1},
                                       "flashCached": True, "growthAlertThreshold": "90",
                                       "expansionPolicy": "automatic", "segmentSize": str(128 * 1024)}
        self.assertEqual(volume_object.get_volume_property_changes(), dict())

        # property changes
        self._set_args(
            {"state": "present", "name": "Matthew", "storage_pool_name": "pool", "size": 100, "ssd_cache_enabled": True,
             "read_cache_enable": True, "write_cache_enable": True,
             "read_ahead_enable": True, "thin_provision": False})
        volume_object = NetAppESeriesVolume()
        volume_object.volume_detail = {"metadata": [],
                                       "cacheSettings": {"readCacheEnable": False, "writeCacheEnable": True,
                                                         "readAheadMultiplier": 1}, "flashCached": True,
                                       "segmentSize": str(128 * 1024)}
        self.assertEqual(volume_object.get_volume_property_changes(),
                         {"metaTags": [], 'cacheSettings': {'readCacheEnable': True, 'writeCacheEnable': True},
                          'flashCache': True})
        self._set_args(
            {"state": "present", "name": "Matthew", "storage_pool_name": "pool", "size": 100, "ssd_cache_enabled": True,
             "read_cache_enable": True, "write_cache_enable": True,
             "read_ahead_enable": True, "thin_provision": False})
        volume_object = NetAppESeriesVolume()
        volume_object.volume_detail = {"metadata": [],
                                       "cacheSettings": {"readCacheEnable": True, "writeCacheEnable": False,
                                                         "readAheadMultiplier": 1}, "flashCached": True,
                                       "segmentSize": str(128 * 1024)}
        self.assertEqual(volume_object.get_volume_property_changes(),
                         {"metaTags": [], 'cacheSettings': {'readCacheEnable': True, 'writeCacheEnable': True},
                          'flashCache': True})
        self._set_args(
            {"state": "present", "name": "Matthew", "storage_pool_name": "pool", "size": 100, "ssd_cache_enabled": True,
             "read_cache_enable": True, "write_cache_enable": True,
             "read_ahead_enable": True, "thin_provision": False})
        volume_object = NetAppESeriesVolume()
        volume_object.volume_detail = {"metadata": [],
                                       "cacheSettings": {"readCacheEnable": True, "writeCacheEnable": True,
                                                         "readAheadMultiplier": 1}, "flashCached": False,
                                       "segmentSize": str(128 * 1024)}
        self.assertEqual(volume_object.get_volume_property_changes(),
                         {"metaTags": [], 'cacheSettings': {'readCacheEnable': True, 'writeCacheEnable': True},
                          'flashCache': True})
        self._set_args(
            {"state": "present", "name": "Matthew", "storage_pool_name": "pool", "size": 100, "ssd_cache_enabled": True,
             "read_cache_enable": True, "write_cache_enable": True,
             "read_ahead_enable": False, "thin_provision": False})
        volume_object = NetAppESeriesVolume()
        volume_object.volume_detail = {"metadata": [],
                                       "cacheSettings": {"readCacheEnable": True, "writeCacheEnable": True,
                                                         "readAheadMultiplier": 1}, "flashCached": False,
                                       "segmentSize": str(128 * 1024)}
        self.assertEqual(volume_object.get_volume_property_changes(), {"metaTags": [],
                                                                       'cacheSettings': {'readCacheEnable': True,
                                                                                         'writeCacheEnable': True,
                                                                                         'readAheadEnable': False},
                                                                       'flashCache': True})

        self._set_args(
            {"state": "present", "name": "Matthew", "storage_pool_name": "pool", "size": 100, "ssd_cache_enabled": True,
             "read_cache_enable": True, "write_cache_enable": True,
             "read_ahead_enable": True, "thin_provision": True, "thin_volume_repo_size": 64,
             "thin_volume_max_repo_size": 1000, "thin_volume_growth_alert_threshold": 90})
        volume_object = NetAppESeriesVolume()
        volume_object.volume_detail = {"metadata": [],
                                       "cacheSettings": {"readCacheEnable": True, "writeCacheEnable": True,
                                                         "readAheadMultiplier": 1},
                                       "flashCached": True, "growthAlertThreshold": "95",
                                       "expansionPolicy": "automatic", "segmentSize": str(128 * 1024)}
        self.assertEqual(volume_object.get_volume_property_changes(),
                         {"metaTags": [], 'cacheSettings': {'readCacheEnable': True, 'writeCacheEnable': True},
                          'growthAlertThreshold': 90, 'flashCache': True})

    def test_get_volume_property_changes_fail(self):
        """Verify correct exception is thrown"""
        self._set_args(
            {"state": "present", "name": "Matthew", "storage_pool_name": "pool", "size": 100, "ssd_cache_enabled": True,
             "read_cache_enable": True, "write_cache_enable": True, "read_ahead_enable": True, "thin_provision": False})
        volume_object = NetAppESeriesVolume()
        volume_object.volume_detail = {
            "cacheSettings": {"readCacheEnable": True, "writeCacheEnable": True, "readAheadMultiplier": 1},
            "flashCached": True, "segmentSize": str(512 * 1024)}
        with self.assertRaisesRegexp(AnsibleFailJson, "Existing volume segment size is"):
            volume_object.get_volume_property_changes()

    def test_get_expand_volume_changes_pass(self):
        """Verify expansion changes."""
        # thick volumes
        self._set_args(
            {"state": "present", "name": "Matthew", "storage_pool_name": "pool", "size": 100, "thin_provision": False})
        volume_object = NetAppESeriesVolume()
        volume_object.volume_detail = {"capacity": str(50 * 1024 * 1024 * 1024), "thinProvisioned": False}
        self.assertEqual(volume_object.get_expand_volume_changes(),
                         {"sizeUnit": "bytes", "expansionSize": 100 * 1024 * 1024 * 1024})

        # thin volumes
        self._set_args(
            {"state": "present", "name": "Matthew", "storage_pool_name": "pool", "size": 100, "thin_provision": True,
             "thin_volume_expansion_policy": "automatic", "thin_volume_repo_size": 64,
             "thin_volume_max_repo_size": 1000, "thin_volume_growth_alert_threshold": 90})
        volume_object = NetAppESeriesVolume()
        volume_object.volume_detail = {"capacity": str(50 * 1024 * 1024 * 1024), "thinProvisioned": True,
                                       "expansionPolicy": "automatic",
                                       "provisionedCapacityQuota": str(1000 * 1024 * 1024 * 1024)}
        self.assertEqual(volume_object.get_expand_volume_changes(),
                         {"sizeUnit": "bytes", "newVirtualSize": 100 * 1024 * 1024 * 1024})
        self._set_args(
            {"state": "present", "name": "Matthew", "storage_pool_name": "pool", "size": 100, "thin_provision": True,
             "thin_volume_expansion_policy": "automatic", "thin_volume_repo_size": 64,
             "thin_volume_max_repo_size": 1000, "thin_volume_growth_alert_threshold": 90})
        volume_object = NetAppESeriesVolume()
        volume_object.volume_detail = {"capacity": str(100 * 1024 * 1024 * 1024), "thinProvisioned": True,
                                       "expansionPolicy": "automatic",
                                       "provisionedCapacityQuota": str(500 * 1024 * 1024 * 1024)}
        self.assertEqual(volume_object.get_expand_volume_changes(),
                         {"sizeUnit": "bytes", "newRepositorySize": 1000 * 1024 * 1024 * 1024})
        self._set_args(
            {"state": "present", "name": "Matthew", "storage_pool_name": "pool", "size": 100, "thin_provision": True,
             "thin_volume_expansion_policy": "manual", "thin_volume_repo_size": 504, "thin_volume_max_repo_size": 1000,
             "thin_volume_growth_alert_threshold": 90})
        volume_object = NetAppESeriesVolume()
        volume_object.volume_detail = {"capacity": str(100 * 1024 * 1024 * 1024), "thinProvisioned": True,
                                       "expansionPolicy": "manual",
                                       "currentProvisionedCapacity": str(500 * 1024 * 1024 * 1024)}
        self.assertEqual(volume_object.get_expand_volume_changes(),
                         {"sizeUnit": "bytes", "newRepositorySize": 504 * 1024 * 1024 * 1024})
        self._set_args(
            {"state": "present", "name": "Matthew", "storage_pool_name": "pool", "size": 100, "thin_provision": True,
             "thin_volume_expansion_policy": "manual", "thin_volume_repo_size": 756, "thin_volume_max_repo_size": 1000,
             "thin_volume_growth_alert_threshold": 90})
        volume_object = NetAppESeriesVolume()
        volume_object.volume_detail = {"capacity": str(100 * 1024 * 1024 * 1024), "thinProvisioned": True,
                                       "expansionPolicy": "manual",
                                       "currentProvisionedCapacity": str(500 * 1024 * 1024 * 1024)}
        self.assertEqual(volume_object.get_expand_volume_changes(),
                         {"sizeUnit": "bytes", "newRepositorySize": 756 * 1024 * 1024 * 1024})

    def test_get_expand_volume_changes_fail(self):
        """Verify exceptions are thrown."""
        self._set_args(
            {"state": "present", "name": "Matthew", "storage_pool_name": "pool", "size": 100, "thin_provision": False})
        volume_object = NetAppESeriesVolume()
        volume_object.volume_detail = {"capacity": str(1000 * 1024 * 1024 * 1024)}
        with self.assertRaisesRegexp(AnsibleFailJson, "Reducing the size of volumes is not permitted."):
            volume_object.get_expand_volume_changes()

        self._set_args(
            {"state": "present", "name": "Matthew", "storage_pool_name": "pool", "size": 100, "thin_provision": True,
             "thin_volume_expansion_policy": "manual", "thin_volume_repo_size": 502, "thin_volume_max_repo_size": 1000,
             "thin_volume_growth_alert_threshold": 90})
        volume_object = NetAppESeriesVolume()
        volume_object.volume_detail = {"capacity": str(100 * 1024 * 1024 * 1024), "thinProvisioned": True,
                                       "expansionPolicy": "manual",
                                       "currentProvisionedCapacity": str(500 * 1024 * 1024 * 1024)}
        with self.assertRaisesRegexp(AnsibleFailJson, "The thin volume repository increase must be between or equal"):
            volume_object.get_expand_volume_changes()

        self._set_args(
            {"state": "present", "name": "Matthew", "storage_pool_name": "pool", "size": 100, "thin_provision": True,
             "thin_volume_expansion_policy": "manual", "thin_volume_repo_size": 760, "thin_volume_max_repo_size": 1000,
             "thin_volume_growth_alert_threshold": 90})
        volume_object = NetAppESeriesVolume()
        volume_object.volume_detail = {"capacity": str(100 * 1024 * 1024 * 1024), "thinProvisioned": True,
                                       "expansionPolicy": "manual",
                                       "currentProvisionedCapacity": str(500 * 1024 * 1024 * 1024)}
        with self.assertRaisesRegexp(AnsibleFailJson, "The thin volume repository increase must be between or equal"):
            volume_object.get_expand_volume_changes()

    def test_create_volume_pass(self):
        """Verify volume creation."""
        self._set_args(
            {"state": "present", "name": "Matthew", "storage_pool_name": "pool", "size": 100, "thin_provision": False})
        volume_object = NetAppESeriesVolume()
        volume_object.pool_detail = {"id": "12345"}
        with mock.patch(self.REQUEST_FUNC, return_value=(200, {})):
            volume_object.create_volume()

        self._set_args(
            {"state": "present", "name": "Matthew", "storage_pool_name": "pool", "size": 100, "thin_provision": True,
             "thin_volume_expansion_policy": "manual", "thin_volume_repo_size": 760, "thin_volume_max_repo_size": 1000,
             "thin_volume_growth_alert_threshold": 90})
        volume_object = NetAppESeriesVolume()
        volume_object.pool_detail = {"id": "12345"}
        with mock.patch(self.REQUEST_FUNC, return_value=(200, {})):
            volume_object.create_volume()

    def test_create_volume_fail(self):
        """Verify exceptions thrown."""
        self._set_args(
            {"state": "present", "name": "Matthew", "storage_pool_name": "pool", "size": 100, "thin_provision": False})
        volume_object = NetAppESeriesVolume()
        volume_object.pool_detail = {"id": "12345"}
        with self.assertRaisesRegexp(AnsibleFailJson, "Failed to create volume."):
            with mock.patch(self.REQUEST_FUNC, return_value=Exception()):
                volume_object.create_volume()

        self._set_args(
            {"state": "present", "name": "Matthew", "storage_pool_name": "pool", "size": 100, "thin_provision": True,
             "thin_volume_expansion_policy": "manual", "thin_volume_repo_size": 760, "thin_volume_max_repo_size": 1000,
             "thin_volume_growth_alert_threshold": 90})
        volume_object = NetAppESeriesVolume()
        volume_object.pool_detail = {"id": "12345"}
        with self.assertRaisesRegexp(AnsibleFailJson, "Failed to create thin volume."):
            with mock.patch(self.REQUEST_FUNC, return_value=Exception()):
                volume_object.create_volume()

    def test_update_volume_properties_pass(self):
        """verify property update."""
        self._set_args(
            {"state": "present", "name": "Matthew", "storage_pool_name": "pool", "size": 100, "thin_provision": False})
        volume_object = NetAppESeriesVolume()
        volume_object.pool_detail = {"id": "12345"}
        volume_object.wait_for_volume_availability = lambda: None
        volume_object.get_volume = lambda: {"id": "12345'"}
        volume_object.get_volume_property_changes = lambda: {
            'cacheSettings': {'readCacheEnable': True, 'writeCacheEnable': True}, 'growthAlertThreshold': 90,
            'flashCached': True}
        volume_object.workload_id = "4200000001000000000000000000000000000000"
        with mock.patch(self.REQUEST_FUNC, return_value=(200, {})):
            self.assertTrue(volume_object.update_volume_properties())

        self._set_args(
            {"state": "present", "name": "Matthew", "storage_pool_name": "pool", "size": 100, "thin_provision": True,
             "thin_volume_expansion_policy": "manual", "thin_volume_repo_size": 760, "thin_volume_max_repo_size": 1000,
             "thin_volume_growth_alert_threshold": 90})
        volume_object = NetAppESeriesVolume()
        volume_object.pool_detail = {"id": "12345"}
        volume_object.wait_for_volume_availability = lambda: None
        volume_object.get_volume = lambda: {"id": "12345'"}
        volume_object.get_volume_property_changes = lambda: {
            'cacheSettings': {'readCacheEnable': True, 'writeCacheEnable': True}, 'growthAlertThreshold': 90,
            'flashCached': True}
        volume_object.workload_id = "4200000001000000000000000000000000000000"
        with mock.patch(self.REQUEST_FUNC, return_value=(200, {})):
            self.assertTrue(volume_object.update_volume_properties())

        self._set_args(
            {"state": "present", "name": "Matthew", "storage_pool_name": "pool", "size": 100, "thin_provision": False})
        volume_object = NetAppESeriesVolume()
        volume_object.pool_detail = {"metadata": [{"key": "workloadId", "value": "12345"}]}
        volume_object.wait_for_volume_availability = lambda: None
        volume_object.get_volume = lambda: {"id": "12345'"}
        volume_object.get_volume_property_changes = lambda: {}
        volume_object.workload_id = "4200000001000000000000000000000000000000"
        self.assertFalse(volume_object.update_volume_properties())

    def test_update_volume_properties_fail(self):
        """Verify exceptions are thrown."""
        self._set_args(
            {"state": "present", "name": "Matthew", "storage_pool_name": "pool", "size": 100, "thin_provision": False})
        volume_object = NetAppESeriesVolume()
        volume_object.pool_detail = {"id": "12345"}
        volume_object.wait_for_volume_availability = lambda: None
        volume_object.get_volume = lambda: {"id": "12345'"}
        volume_object.get_volume_property_changes = lambda: {
            'cacheSettings': {'readCacheEnable': True, 'writeCacheEnable': True}, 'growthAlertThreshold': 90,
            'flashCached': True}
        volume_object.workload_id = "4200000001000000000000000000000000000000"
        with self.assertRaisesRegexp(AnsibleFailJson, "Failed to update volume properties."):
            with mock.patch(self.REQUEST_FUNC, return_value=Exception()):
                self.assertTrue(volume_object.update_volume_properties())

        self._set_args(
            {"state": "present", "name": "Matthew", "storage_pool_name": "pool", "size": 100, "thin_provision": True,
             "thin_volume_expansion_policy": "manual", "thin_volume_repo_size": 760, "thin_volume_max_repo_size": 1000,
             "thin_volume_growth_alert_threshold": 90})
        volume_object = NetAppESeriesVolume()
        volume_object.pool_detail = {"id": "12345"}
        volume_object.wait_for_volume_availability = lambda: None
        volume_object.get_volume = lambda: {"id": "12345'"}
        volume_object.get_volume_property_changes = lambda: {
            'cacheSettings': {'readCacheEnable': True, 'writeCacheEnable': True}, 'growthAlertThreshold': 90,
            'flashCached': True}
        volume_object.workload_id = "4200000001000000000000000000000000000000"
        with self.assertRaisesRegexp(AnsibleFailJson, "Failed to update thin volume properties."):
            with mock.patch(self.REQUEST_FUNC, return_value=Exception()):
                self.assertTrue(volume_object.update_volume_properties())

    def test_expand_volume_pass(self):
        """Verify volume expansion."""
        self._set_args(
            {"state": "present", "name": "Matthew", "storage_pool_name": "pool", "size": 100, "thin_provision": False})
        volume_object = NetAppESeriesVolume()
        volume_object.get_expand_volume_changes = lambda: {"sizeUnit": "bytes",
                                                           "expansionSize": 100 * 1024 * 1024 * 1024}
        volume_object.volume_detail = {"id": "12345", "thinProvisioned": True}
        with mock.patch(self.REQUEST_FUNC, return_value=(200, {})):
            volume_object.expand_volume()

        self._set_args(
            {"state": "present", "name": "Matthew", "storage_pool_name": "pool", "size": 100, "thin_provision": True,
             "thin_volume_expansion_policy": "manual", "thin_volume_repo_size": 760, "thin_volume_max_repo_size": 1000,
             "thin_volume_growth_alert_threshold": 90})
        volume_object = NetAppESeriesVolume()
        volume_object.get_expand_volume_changes = lambda: {"sizeUnit": "bytes",
                                                           "expansionSize": 100 * 1024 * 1024 * 1024}
        volume_object.volume_detail = {"id": "12345", "thinProvisioned": True}
        with mock.patch(self.REQUEST_FUNC, return_value=(200, {})):
            volume_object.expand_volume()

    def test_expand_volume_fail(self):
        """Verify exceptions are thrown."""
        self._set_args(
            {"state": "present", "name": "Matthew", "storage_pool_name": "pool", "size": 100, "thin_provision": False})
        volume_object = NetAppESeriesVolume()
        volume_object.get_expand_volume_changes = lambda: {"sizeUnit": "bytes",
                                                           "expansionSize": 100 * 1024 * 1024 * 1024}
        volume_object.volume_detail = {"id": "12345", "thinProvisioned": False}
        with self.assertRaisesRegexp(AnsibleFailJson, "Failed to expand volume."):
            with mock.patch(self.REQUEST_FUNC, return_value=Exception()):
                volume_object.expand_volume()

        self._set_args(
            {"state": "present", "name": "Matthew", "storage_pool_name": "pool", "size": 100, "thin_provision": True})
        volume_object = NetAppESeriesVolume()
        volume_object.get_expand_volume_changes = lambda: {"sizeUnit": "bytes",
                                                           "expansionSize": 100 * 1024 * 1024 * 1024}
        volume_object.volume_detail = {"id": "12345", "thinProvisioned": True}
        with self.assertRaisesRegexp(AnsibleFailJson, "Failed to expand thin volume."):
            with mock.patch(self.REQUEST_FUNC, return_value=Exception()):
                volume_object.expand_volume()

    def test_delete_volume_pass(self):
        """Verify volume deletion."""
        self._set_args(
            {"state": "present", "name": "Matthew", "storage_pool_name": "pool", "size": 100, "thin_provision": False})
        volume_object = NetAppESeriesVolume()
        volume_object.volume_detail = {"id": "12345"}
        with mock.patch(self.REQUEST_FUNC, return_value=(200, {})):
            volume_object.delete_volume()

        self._set_args(
            {"state": "present", "name": "Matthew", "storage_pool_name": "pool", "size": 100, "thin_provision": True,
             "thin_volume_expansion_policy": "manual", "thin_volume_repo_size": 760, "thin_volume_max_repo_size": 1000,
             "thin_volume_growth_alert_threshold": 90})
        volume_object = NetAppESeriesVolume()
        volume_object.volume_detail = {"id": "12345"}
        with mock.patch(self.REQUEST_FUNC, return_value=(200, {})):
            volume_object.delete_volume()

    def test_delete_volume_fail(self):
        """Verify exceptions are thrown."""
        self._set_args(
            {"state": "present", "name": "Matthew", "storage_pool_name": "pool", "size": 100, "thin_provision": False})
        volume_object = NetAppESeriesVolume()
        with self.assertRaisesRegexp(AnsibleFailJson, "Failed to delete volume."):
            with mock.patch(self.REQUEST_FUNC, return_value=Exception()):
                volume_object.delete_volume()

        self._set_args(
            {"state": "present", "name": "Matthew", "storage_pool_name": "pool", "size": 100, "thin_provision": True})
        volume_object = NetAppESeriesVolume()
        with self.assertRaisesRegexp(AnsibleFailJson, "Failed to delete thin volume."):
            with mock.patch(self.REQUEST_FUNC, return_value=Exception()):
                volume_object.delete_volume()



# (c) 2018 Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

import base64
import json
import os
import pkgutil
import random
import re

from distutils.version import LooseVersion

from ansible import constants as C
from ansible.errors import AnsibleError
from ansible.module_utils._text import to_bytes, to_text
from ansible.plugins.loader import ps_module_utils_loader


class PSModuleDepFinder(object):

    def __init__(self):
        self.ps_modules = dict()
        self.exec_scripts = dict()

        # by defining an explicit dict of cs utils and where they are used, we
        # can potentially save time by not adding the type multiple times if it
        # isn't needed
        self.cs_utils_wrapper = dict()
        self.cs_utils_module = dict()

        self.ps_version = None
        self.os_version = None
        self.become = False

        self._re_cs_module = re.compile(to_bytes(r'(?i)^using\s((Ansible\..+)|(AnsibleCollections\.\w.+\.\w.+\w.+));\s*$'))
        self._re_cs_in_ps_module = re.compile(to_bytes(r'(?i)^#\s*ansiblerequires\s+-csharputil\s+((Ansible\..+)|(AnsibleCollections\.\w.+\.\w.+\w.+))'))
        self._re_coll_ps_in_ps_module = re.compile(to_bytes(r'(?i)^#\s*ansiblerequires\s+-powershell\s+((Ansible\..+)|(AnsibleCollections\.\w.+\.\w.+\w.+))'))
        self._re_module = re.compile(to_bytes(r'(?i)^#\s*requires\s+\-module(?:s?)\s*(Ansible\.ModuleUtils\..+)'))
        self._re_wrapper = re.compile(to_bytes(r'(?i)^#\s*ansiblerequires\s+-wrapper\s+(\w*)'))
        self._re_ps_version = re.compile(to_bytes(r'(?i)^#requires\s+\-version\s+([0-9]+(\.[0-9]+){0,3})$'))
        self._re_os_version = re.compile(to_bytes(r'(?i)^#ansiblerequires\s+\-osversion\s+([0-9]+(\.[0-9]+){0,3})$'))
        self._re_become = re.compile(to_bytes(r'(?i)^#ansiblerequires\s+\-become$'))

    def scan_module(self, module_data, wrapper=False, powershell=True):
        lines = module_data.split(b'\n')
        module_utils = set()
        if wrapper:
            cs_utils = self.cs_utils_wrapper
        else:
            cs_utils = self.cs_utils_module

        if powershell:
            checks = [
                # PS module contains '#Requires -Module Ansible.ModuleUtils.*'
                (self._re_module, self.ps_modules, ".psm1"),
                # PS module contains '#AnsibleRequires -Powershell Ansible.*' (or FQ collections module_utils ref)
                (self._re_coll_ps_in_ps_module, self.ps_modules, ".psm1"),
                # PS module contains '#AnsibleRequires -CSharpUtil Ansible.*'
                (self._re_cs_in_ps_module, cs_utils, ".cs"),
            ]
        else:
            checks = [
                # CS module contains 'using Ansible.*;' or 'using AnsibleCollections.ns.coll.*;'
                (self._re_cs_module, cs_utils, ".cs"),
            ]

        for line in lines:
            for check in checks:
                match = check[0].match(line)
                if match:
                    # tolerate windows line endings by stripping any remaining
                    # newline chars
                    module_util_name = self._normalize_mu_name(match.group(1).rstrip())

                    if module_util_name not in check[1].keys():
                        module_utils.add((module_util_name, check[2]))

            if powershell:
                ps_version_match = self._re_ps_version.match(line)
                if ps_version_match:
                    self._parse_version_match(ps_version_match, "ps_version")

                os_version_match = self._re_os_version.match(line)
                if os_version_match:
                    self._parse_version_match(os_version_match, "os_version")

                # once become is set, no need to keep on checking recursively
                if not self.become:
                    become_match = self._re_become.match(line)
                    if become_match:
                        self.become = True

            if wrapper:
                wrapper_match = self._re_wrapper.match(line)
                if wrapper_match:
                    self.scan_exec_script(wrapper_match.group(1).rstrip())

        # recursively drill into each Requires to see if there are any more
        # requirements
        for m in set(module_utils):
            self._add_module(m, wrapper=wrapper)

    def scan_exec_script(self, name):
        # scans lib/ansible/executor/powershell for scripts used in the module
        # exec side. It also scans these scripts for any dependencies
        name = to_text(name)
        if name in self.exec_scripts.keys():
            return

        data = pkgutil.get_data("ansible.executor.powershell", name + ".ps1")
        if data is None:
            raise AnsibleError("Could not find executor powershell script "
                               "for '%s'" % name)

        b_data = to_bytes(data)

        # remove comments to reduce the payload size in the exec wrappers
        if C.DEFAULT_DEBUG:
            exec_script = b_data
        else:
            exec_script = _strip_comments(b_data)
        self.exec_scripts[name] = to_bytes(exec_script)
        self.scan_module(b_data, wrapper=True, powershell=True)

    def _add_module(self, name, wrapper=False):
        m, ext = name
        m = to_text(m)
        mu_path = ps_module_utils_loader.find_plugin(m, ext)
        if not mu_path:
            raise AnsibleError('Could not find imported module support code '
                               'for \'%s\'' % m)

        module_util_data = to_bytes(_slurp(mu_path))
        if ext == ".psm1":
            self.ps_modules[m] = module_util_data
        else:
            if wrapper:
                self.cs_utils_wrapper[m] = module_util_data
            else:
                self.cs_utils_module[m] = module_util_data
        self.scan_module(module_util_data, wrapper=wrapper,
                         powershell=(ext == ".psm1"))

    def _parse_version_match(self, match, attribute):
        new_version = to_text(match.group(1)).rstrip()

        # PowerShell cannot cast a string of "1" to Version, it must have at
        # least the major.minor for it to be valid so we append 0
        if match.group(2) is None:
            new_version = "%s.0" % new_version

        existing_version = getattr(self, attribute, None)
        if existing_version is None:
            setattr(self, attribute, new_version)
        else:
            # determine which is the latest version and set that
            if LooseVersion(new_version) > LooseVersion(existing_version):
                setattr(self, attribute, new_version)

    def _normalize_mu_name(self, mu):
        # normalize Windows module_utils to remove 'AnsibleCollections.' prefix so the plugin loader can find them
        mu = to_text(mu)

        if not mu.startswith(u'AnsibleCollections.'):
            return mu

        return mu.replace(u'AnsibleCollections.', u'', 1)


def _slurp(path):
    if not os.path.exists(path):
        raise AnsibleError("imported module support code does not exist at %s"
                           % os.path.abspath(path))
    fd = open(path, 'rb')
    data = fd.read()
    fd.close()
    return data


def _strip_comments(source):
    # Strip comments and blank lines from the wrapper
    buf = []
    start_block = False
    for line in source.splitlines():
        l = line.strip()

        if start_block and l.endswith(b'#>'):
            start_block = False
            continue
        elif start_block:
            continue
        elif l.startswith(b'<#'):
            start_block = True
            continue
        elif not l or l.startswith(b'#'):
            continue

        buf.append(line)
    return b'\n'.join(buf)


def _create_powershell_wrapper(b_module_data, module_args, environment,
                               async_timeout, become, become_method,
                               become_user, become_password, become_flags,
                               substyle):
    # creates the manifest/wrapper used in PowerShell/C# modules to enable
    # things like become and async - this is also called in action/script.py

    # FUTURE: add process_wrapper.ps1 to run module_wrapper in a new process
    # if running under a persistent connection and substyle is C# so we
    # don't have type conflicts
    finder = PSModuleDepFinder()
    if substyle != 'script':
        # don't scan the module for util dependencies and other Ansible related
        # flags if the substyle is 'script' which is set by action/script
        finder.scan_module(b_module_data, powershell=(substyle == "powershell"))

    module_wrapper = "module_%s_wrapper" % substyle
    exec_manifest = dict(
        module_entry=to_text(base64.b64encode(b_module_data)),
        powershell_modules=dict(),
        csharp_utils=dict(),
        csharp_utils_module=list(),  # csharp_utils only required by a module
        module_args=module_args,
        actions=[module_wrapper],
        environment=environment,
        encoded_output=False
    )
    finder.scan_exec_script(module_wrapper)

    if async_timeout > 0:
        finder.scan_exec_script('exec_wrapper')
        finder.scan_exec_script('async_watchdog')
        finder.scan_exec_script('async_wrapper')

        exec_manifest["actions"].insert(0, 'async_watchdog')
        exec_manifest["actions"].insert(0, 'async_wrapper')
        exec_manifest["async_jid"] = str(random.randint(0, 999999999999))
        exec_manifest["async_timeout_sec"] = async_timeout

    if become and become_method == 'runas':
        finder.scan_exec_script('exec_wrapper')
        finder.scan_exec_script('become_wrapper')

        exec_manifest["actions"].insert(0, 'become_wrapper')
        exec_manifest["become_user"] = become_user
        exec_manifest["become_password"] = become_password
        exec_manifest['become_flags'] = become_flags

    exec_manifest['min_ps_version'] = finder.ps_version
    exec_manifest['min_os_version'] = finder.os_version
    if finder.become and 'become_wrapper' not in exec_manifest['actions']:
        finder.scan_exec_script('exec_wrapper')
        finder.scan_exec_script('become_wrapper')

        exec_manifest['actions'].insert(0, 'become_wrapper')
        exec_manifest['become_user'] = 'SYSTEM'
        exec_manifest['become_password'] = None
        exec_manifest['become_flags'] = None

    # make sure Ansible.ModuleUtils.AddType is added if any C# utils are used
    if len(finder.cs_utils_wrapper) > 0 or len(finder.cs_utils_module) > 0:
        finder._add_module((b"Ansible.ModuleUtils.AddType", ".psm1"),
                           wrapper=False)

    # exec_wrapper is only required to be part of the payload if using
    # become or async, to save on payload space we check if exec_wrapper has
    # already been added, and remove it manually if it hasn't later
    exec_required = "exec_wrapper" in finder.exec_scripts.keys()
    finder.scan_exec_script("exec_wrapper")
    # must contain an empty newline so it runs the begin/process/end block
    finder.exec_scripts["exec_wrapper"] += b"\n\n"

    exec_wrapper = finder.exec_scripts["exec_wrapper"]
    if not exec_required:
        finder.exec_scripts.pop("exec_wrapper")

    for name, data in finder.exec_scripts.items():
        b64_data = to_text(base64.b64encode(data))
        exec_manifest[name] = b64_data

    for name, data in finder.ps_modules.items():
        b64_data = to_text(base64.b64encode(data))
        exec_manifest['powershell_modules'][name] = b64_data

    cs_utils = finder.cs_utils_wrapper
    cs_utils.update(finder.cs_utils_module)
    for name, data in cs_utils.items():
        b64_data = to_text(base64.b64encode(data))
        exec_manifest['csharp_utils'][name] = b64_data
    exec_manifest['csharp_utils_module'] = list(finder.cs_utils_module.keys())

    b_json = to_bytes(json.dumps(exec_manifest))
    # delimit the payload JSON from the wrapper to keep sensitive contents out of scriptblocks (which can be logged)
    b_data = exec_wrapper + b'\0\0\0\0' + b_json
    return b_data
